/*		Structured stream to Rich hypertext converter
**		============================================
**
**	This generates of a hypertext object.  It converts from the
**	structured stream interface fro HTMl events into the style-
**	oriented iunterface of the HText.h interface.  This module is
**	only used in clients and shouldnot be linked into servers.
**
**	Override this module if making a new GUI browser. Maybe.
**
*/

#include <appkit/appkit.h>	/* Seems to need to be first */

#include "HTML.h"

/* #define CAREFUL		 Check nesting here notreally necessary */

#include <ctype.h>
#include <stdio.h>

#include "HTAtom.h"
#include "HTChunk.h"
#include "HText.h"
#include "HTStyle.h"
#include "HTAlert.h"
#include "HTMLGen.h"
#include "HTParse.h"
#include "SGML.h"

#ifdef NeXT
PRIVATE char bullets[] = "\267*o+-.*o+-.*o+-.";
#else
PRIVATE char bullets[] = "*o+-.*o+-.*o+-.";
#endif

extern HTStyleSheet * styleSheet;	/* Application-wide */

/*	Module-wide style cache
*/
PRIVATE int 		got_styles = 0;
PRIVATE HTStyle *styles[HTMLP_ELEMENTS];
PRIVATE HTStyle *default_style;


#define NESTING_MAGIC_NUMBER 7763129

/*		SGML Object
**		-----------
**
**	This object optionally stores the whole structued object,
**	rather than just passing it on as style info.
*/

/*	Extra information about SGML elements
**	-------------------------------------
*/

typedef struct _SGML_extra {
    char *	paragraph_tag;	/* Tag to mark paragraph mark within style */
    char *	tab_tag;	/* Tag to mark tab within style 	*/

    char *	start_text;	/* Text conventionally starting this style */
    char *	paragraph_text;	/* Text used as a paragraph mark within style*/
    char *	end_text;	/* Text used to end a style 		*/
    int		free_format;	/* Flag: are line ends word breaks only? */
} SGML_extra;


/*	Extra information about SGML elements
**	-------------------------------------

There is a small number of different ways of treating white
space in SGML, in mapping from the text onject to HTML.
These have to be programmed it seems.
	
In text object	\n\n		\n	tab	\n\n\t
--------------	-------------	-----	-----	-------
in Address,
Blockquote,
Normal,		<P>		<BR>	-		NORMAL
H1-6:		close+open	<BR>	-		HEADING
Glossary	<DT>		<DT>	<DD>	<P>	GLOSSARY
List,				
Menu		<LI>		<LI>	-	<P>	LIST
Dir		<LI>		<LI>	<LI>		DIR
Pre etc		\n\n		\n	\t		PRE
**
*/

typedef enum _white_space_treatment {
	WS_NORMAL,
	WS_HEADING,
	WS_GLOSSARY,
	WS_LIST,
	WS_DIR,
	WS_PRE
} white_space_treatment;


/*	A tree with an item for each nesting state.
**	------
**
**	The first element must be the style so that we can easily
**	from the style pointer get the nesting state.
*/
typedef struct _nesting_state {
    HTStyle			style;	/* @@@@ */
    white_space_treatment	wst;
    struct _nesting_state *	parent;
    /* HTTag * 			tag; */
    int				element_number;
    int				item_number;	/* only for ordered lists */
    int				list_level;	/* how deep nested */
    HTList *			children;
    BOOL			paragraph_break;
    int				magic;
    BOOL			object_gens_HTML; /* we don't generate HTML */
} nesting_state;


/*	The document contents tree
**	--------------------------
*/

enum item_type { HT_TEXT, HT_IMAGE, HT_ELEMENT, HT_DOCUMENT };
typedef struct _document_item {
    enum item_type 	itype;
    struct _document_item * parent;	/* enclosing item */
    nesting_state *	nesting;	/* SGML nesting for this item */
    int			total_chars;	/* Characters in this item */
    int			total_lines;	/* Lines in this item */
    union {
    	struct {
	    int 	dummy;
	} document;
	struct {
	    HTChunk *	data;	    
	} text;
	struct {
	    HTList * 	contents;
	} element;
	struct {
	    HTParentAnchor * anchor;
	} image;
    } as;
} document_item;

#define MAX_LIST_DEPTH 6;


/*	The document
**	------------
*/
struct _HTStructured {
    CONST HTStructuredClass * 	isa;
    HTParentAnchor * 		node_anchor;
    HText * 			text;

    BOOL			store;			/* Do we store doc? */
    
    CONST SGML_dtd*		dtd;			/* Document type def */
    
    HTStream*			target;			/* Output stream */
    HTStreamClass		targetClass;		/* Output routines */

    HTChunk 			title;		/* Grow by 128 */
    
    char *			comment_start;	/* for literate programming */
    char *			comment_end;

    HTTag *			current_tag;
    BOOL			style_change;
    HTStyle *			new_style;
    BOOL			para_break_on_new_style;
    HTStyle *			old_style;
    BOOL			in_word;  /* Have just had a non-white char */
    
    document_item*		current_item;	/* Tree node */
    nesting_state*		current_nesting;	/* Tree node */

    document_item		document;	/* Tree root */
    
    nesting_state		nest;		/* Tree root */
    
    int				reference_count; /* for freeing */
    HTRequest *			request;	/* for passing on only */
    
    BOOL			build_finished;
};

struct _HTStream {
    CONST HTStreamClass *	isa;
    /* .... */
};

#define PUTC(c) (*target->isa->put_character)(target, c)
#define PUTS(s) (*target->isa->put_string)(target, s)
#define START(e) (*target->isa->start_element)(target, e, 0, 0)
#define END(e) (*target->isa->end_element)(target, e)
#define END_TARGET (*target->isa->end_document)(target)
#define FREE_TARGET (*target->isa->free)(target)


/*______________________________________________________________________
*/

/*		Forward declarations of routines
*/
PRIVATE void get_styles NOPARAMS;

PRIVATE void HTStyleModify PARAMS((nesting_state* nesting,
		int element_number));

PRIVATE void actually_set_style PARAMS((HTStructured * me));

/*	Style buffering avoids dummy paragraph begin/ends.
*/
#define UPDATE_STYLE if (me->style_change) { actually_set_style(me); }


#ifdef OLD_CODE
/* The following accented characters are from peter Flynn, curia project */

/* these ifdefs don't solve the problem of a simple terminal emulator
** with a different character set to the client machine. But nothing does,
** except looking at the TERM setting */


        { "ocus" , "&" },       /* for CURIA */
#ifdef IBMPC
        { "aacute" , "\240" },	/* For PC display */
        { "eacute" , "\202" },
        { "iacute" , "\241" },
        { "oacute" , "\242" },
        { "uacute" , "\243" },
        { "Aacute" , "\101" },
        { "Eacute" , "\220" },
        { "Iacute" , "\111" },
        { "Oacute" , "\117" },
        { "Uacute" , "\125" },
#else
        { "aacute" , "\341" },	/* Works for openwindows -- Peter Flynn */
        { "eacute" , "\351" },
        { "iacute" , "\355" },
        { "oacute" , "\363" },
        { "uacute" , "\372" },
        { "Aacute" , "\301" },
        { "Eacute" , "\310" },
        { "Iacute" , "\315" },
        { "Oacute" , "\323" },
        { "Uacute" , "\332" }, 
#endif
	{ 0,	0 }  /* Terminate list */
};
#endif


/* 	Entity values -- for ISO Latin 1 local representation
**
**	This MUST match exactly the table referred to in the DTD!
*/
static char * ISO_Latin1[] = {
  	"\306",	/* capital AE diphthong (ligature) */ 
  	"\301",	/* capital A, acute accent */ 
  	"\302",	/* capital A, circumflex accent */ 
  	"\300",	/* capital A, grave accent */ 
  	"\305",	/* capital A, ring */ 
  	"\303",	/* capital A, tilde */ 
  	"\304",	/* capital A, dieresis or umlaut mark */ 
  	"\307",	/* capital C, cedilla */ 
  	"\320",	/* capital Eth, Icelandic */ 
  	"\311",	/* capital E, acute accent */ 
  	"\312",	/* capital E, circumflex accent */ 
  	"\310",	/* capital E, grave accent */ 
  	"\313",	/* capital E, dieresis or umlaut mark */ 
  	"\315",	/* capital I, acute accent */ 
  	"\316",	/* capital I, circumflex accent */ 
  	"\314",	/* capital I, grave accent */ 
  	"\317",	/* capital I, dieresis or umlaut mark */ 
  	"\321",	/* capital N, tilde */ 
  	"\323",	/* capital O, acute accent */ 
  	"\324",	/* capital O, circumflex accent */ 
  	"\322",	/* capital O, grave accent */ 
  	"\330",	/* capital O, slash */ 
  	"\325",	/* capital O, tilde */ 
  	"\326",	/* capital O, dieresis or umlaut mark */ 
  	"\336",	/* capital THORN, Icelandic */ 
  	"\332",	/* capital U, acute accent */ 
  	"\333",	/* capital U, circumflex accent */ 
  	"\331",	/* capital U, grave accent */ 
  	"\334",	/* capital U, dieresis or umlaut mark */ 
  	"\335",	/* capital Y, acute accent */ 
  	"\341",	/* small a, acute accent */ 
  	"\342",	/* small a, circumflex accent */ 
  	"\346",	/* small ae diphthong (ligature) */ 
  	"\340",	/* small a, grave accent */ 
  	"\046",	/* ampersand */ 
  	"\345",	/* small a, ring */ 
  	"\343",	/* small a, tilde */ 
  	"\344",	/* small a, dieresis or umlaut mark */ 
  	"\347",	/* small c, cedilla */ 
  	"\351",	/* small e, acute accent */ 
  	"\352",	/* small e, circumflex accent */ 
  	"\350",	/* small e, grave accent */ 
  	"\360",	/* small eth, Icelandic */ 
  	"\353",	/* small e, dieresis or umlaut mark */ 
  	"\076",	/* greater than */ 
  	"\355",	/* small i, acute accent */ 
  	"\356",	/* small i, circumflex accent */ 
  	"\354",	/* small i, grave accent */ 
  	"\357",	/* small i, dieresis or umlaut mark */ 
  	"\074",	/* less than */ 
  	"\361",	/* small n, tilde */ 
  	"\363",	/* small o, acute accent */ 
  	"\364",	/* small o, circumflex accent */ 
  	"\362",	/* small o, grave accent */ 
  	"\370",	/* small o, slash */ 
  	"\365",	/* small o, tilde */ 
  	"\366",	/* small o, dieresis or umlaut mark */ 
  	"\337",	/* small sharp s, German (sz ligature) */ 
  	"\376",	/* small thorn, Icelandic */ 
  	"\372",	/* small u, acute accent */ 
  	"\373",	/* small u, circumflex accent */ 
  	"\371",	/* small u, grave accent */ 
  	"\374",	/* small u, dieresis or umlaut mark */ 
  	"\375",	/* small y, acute accent */ 
  	"\377",	/* small y, dieresis or umlaut mark */ 
};


/* 	Entity values -- for NeXT local representation
**
**	This MUST match exactly the table referred to in the DTD!
**
*/
static char * NeXTCharacters[] = {
  	"\341",	/* capital AE diphthong (ligature) 	*/ 
  	"\202",	/* capital A, acute accent		*/ 
  	"\203",	/* capital A, circumflex accent 	*/ 
  	"\201",	/* capital A, grave accent 		*/ 
  	"\206",	/* capital A, ring 			*/ 
  	"\204",	/* capital A, tilde 			*/ 
  	"\205",	/* capital A, dieresis or umlaut mark	*/ 
  	"\207",	/* capital C, cedilla 			*/ 
  	"\220",	/* capital Eth, Icelandic 		*/ 
  	"\211",	/* capital E, acute accent 				*/ 
  	"\212",	/* capital E, circumflex accent 			*/ 
  	"\210",	/* capital E, grave accent 				*/ 
  	"\213",	/* capital E, dieresis or umlaut mark 			*/ 
  	"\215",	/* capital I, acute accent 				*/ 
  	"\216",	/* capital I, circumflex accent 	these are	*/ 
  	"\214",	/* capital I, grave accent		ISO -100 hex	*/ 
  	"\217",	/* capital I, dieresis or umlaut mark			*/ 
  	"\221",	/* capital N, tilde 					*/ 
  	"\223",	/* capital O, acute accent 				*/ 
  	"\224",	/* capital O, circumflex accent 			*/ 
  	"\222",	/* capital O, grave accent 				*/ 
  	"\351",	/* capital O, slash 		'cept this */ 
  	"\225",	/* capital O, tilde 					*/ 
  	"\226",	/* capital O, dieresis or umlaut mark			*/ 
  	"\234",	/* capital THORN, Icelandic */ 
  	"\230",	/* capital U, acute accent */ 
  	"\231",	/* capital U, circumflex accent */ 
  	"\227",	/* capital U, grave accent */ 
  	"\232",	/* capital U, dieresis or umlaut mark */ 
  	"\233",	/* capital Y, acute accent */ 
  	"\326",	/* small a, acute accent */ 
  	"\327",	/* small a, circumflex accent */ 
  	"\361",	/* small ae diphthong (ligature) */ 
  	"\325",	/* small a, grave accent */ 
  	"\046",	/* ampersand */ 
  	"\332",	/* small a, ring */ 
  	"\330",	/* small a, tilde */ 
  	"\331",	/* small a, dieresis or umlaut mark */ 
  	"\333",	/* small c, cedilla */ 
  	"\335",	/* small e, acute accent */ 
  	"\336",	/* small e, circumflex accent */ 
  	"\334",	/* small e, grave accent */ 
  	"\346",	/* small eth, Icelandic 	*/ 
  	"\337",	/* small e, dieresis or umlaut mark */ 
  	"\076",	/* greater than */ 
  	"\342",	/* small i, acute accent */ 
  	"\344",	/* small i, circumflex accent */ 
  	"\340",	/* small i, grave accent */ 
  	"\345",	/* small i, dieresis or umlaut mark */ 
  	"\074",	/* less than */ 
  	"\347",	/* small n, tilde */ 
  	"\355",	/* small o, acute accent */ 
  	"\356",	/* small o, circumflex accent */ 
  	"\354",	/* small o, grave accent */ 
  	"\371",	/* small o, slash */ 
  	"\357",	/* small o, tilde */ 
  	"\360",	/* small o, dieresis or umlaut mark */ 
  	"\373",	/* small sharp s, German (sz ligature) */ 
  	"\374",	/* small thorn, Icelandic */ 
  	"\363",	/* small u, acute accent */ 
  	"\364",	/* small u, circumflex accent */ 
  	"\362",	/* small u, grave accent */ 
  	"\366",	/* small u, dieresis or umlaut mark */ 
  	"\367",	/* small y, acute accent */ 
  	"\375",	/* small y, dieresis or umlaut mark */ 
};

/* 	Entity values -- for IBM/PC Code Page 850 (International)
**
**	This MUST match exactly the table referred to in the DTD!
**
*/
/* @@@@@@@@@@@@@@@@@ TBD */



/*		Set character set
**		----------------
*/

PRIVATE char** p_entity_values = ISO_Latin1;	/* Pointer to translation */

PUBLIC void HTMLUseCharacterSet ARGS1(HTMLCharacterSet, i)
{
    p_entity_values = (i == HTML_NEXT_CHARS) ? NeXTCharacters
    					     : ISO_Latin1;
}
/*________________________________________________________________________
**
**	Non-HTML-specific style routines
*/

/*		Flattening the style structure
**		------------------------------
**
On the NeXT, and on any read-only browser, it is simpler for the text to have
a sequence of styles, rather than a nested tree of styles. In this
case we have to flatten the structure as it arrives from SGML tags into
a sequence of styles.

The styles are in fact nesting_state objects, so we can
work back to the tree-structure.

*/

/*		If style really needs to be set, call this
*/
PRIVATE void actually_set_style ARGS1(HTStructured *, me)
{
    if (!me->text) {			/* First time through */
	    me->text = HText_new3(me->request, me->target, me);
	    HText_beginAppend(me->text);
	    HText_setStyle(me->text, me->new_style);
	    me->in_word = NO;
    } else {
    	    if (me->para_break_on_new_style &&
	        ((nesting_state*)(me->old_style))->wst != WS_PRE)
		     HText_appendParagraph(me->text);
	    HText_setStyle(me->text, me->new_style);
    }
    me->old_style = me->new_style;
    me->para_break_on_new_style = NO;
    me->style_change = NO;
}

/*      If you THINK you need to change style, call this
*/

PRIVATE void change_paragraph_style ARGS2(HTStructured *, me, HTStyle *,style)
{
    if (me->new_style!=style) {
    	me->style_change = YES;
	me->new_style = style;
    }
    me->in_word = NO;
}

/*	Print out nesting level
*/
PRIVATE void print_nesting(nesting_state * p)
{
    for (;p; p= p->parent) {
/*	fprintf(stderr, " in %s", me->dtd->tags[p->element_number].name); */
	fprintf(stderr, " in %d", p->element_number);
    }
}

/*_________________________________________________________________
**
**			HTML regeneration
**			=================
*/

static char * prefix;		/* Pointer to prefix string to be junked */
static int SGML_gen_newlines;	/* Yukky statics @@@ */
static BOOL stripping;

PUBLIC void HTRegenInit NOARGS
{
    prefix = NULL;
    SGML_gen_newlines = 0;
    stripping = NO;
}


PRIVATE void start_anchor ARGS3(HTStructured *, s,
		HTChildAnchor*, a, CONST char *, saveName)
{
    HTAnchor * da = HTAnchor_followMainLink((HTAnchor*)a);
    HTParentAnchor * pa = HTAnchor_parent((HTAnchor*)a);
    char * name = HTAnchor_address((HTAnchor*)a);
    char * href = NULL;

    if (da) {
	HTParentAnchor * pda = HTAnchor_parent(da);
	char * absolute = HTAnchor_address(da);
	pda = pda ? pda : (HTParentAnchor *)da;		/* The node anchor */
	if (pda == pa) {
	    char * temp = HTParse(absolute, "", PARSE_ANCHOR);
	    StrAllocCopy(href, "#");
	    StrAllocCat(href, temp);
	    free(temp);
	} else {
	    href = HTRelative(absolute, saveName);
	}
	free(absolute);
    }
    HTStartAnchor(s, strchr(name, '#')+1 , href);
    free(name);
    free(href);
}


PUBLIC int HTMLLevel ARGS1(HTStyle *, s)
{
    nesting_state * p;
    int level =0;
    for(p = (nesting_state*)s; p; p = p->parent)
	    level++;
    return level;
}


/*	Find commonality of two nesting states
**	--------------------------------------
**
**	This find teh youngest common ancenstor of two nesting
**	states --ie their greatest degree of commonality.
*/

PUBLIC HTStyle* HTMLCommonality ARGS2( HTStyle *, old, HTStyle *, new)
{
    int old_level, new_level, yca_level;
    nesting_state * old_nesting = (nesting_state*) old;
    nesting_state * new_nesting = (nesting_state*) new;
    nesting_state * yca;	/* youngest_common_ancestor */ ;
    
    old_level = HTMLLevel(old);
    new_level = HTMLLevel(new);
    if(TRACE) {
	    fprintf(stderr, "Was ");
	    print_nesting(old_nesting);
	    fprintf(stderr, ", will be ");
	    print_nesting(new_nesting);
	    fprintf(stderr, ".\n");
    }

    yca=new_nesting;
    
    /* Find equivalent point in new stack if deeper than old */
    for (yca_level=new_level; yca_level > old_level; yca_level--) {
	yca = yca->parent;
    }

    /* Drop levels to get to save level of nesting as new */
    for (; old_nesting != yca; old_level--) {
	old_nesting = old_nesting->parent;
	if (old_level <= yca_level) {
	    yca = yca->parent;	/* move down together */
	}
    }
    
    /* yca and old_nesting now point to Y.C.A */
    return (HTStyle*) yca;
}


/*	Regenerate the structure from the styles
**	----------------------------------------
**
**	This routine takes the shortest path between two
**	nesting states.
**	A recursive implementation would be neater but probably
**	slower.
**
**	We artificially suppress the opening and closing
**	of the dummy outer level (-1) and the wrapping <HTML> element.
*/

PUBLIC	void HTNestingChange ARGS5
	(HTStructured*, 	s, 
	HTStyle*, 		old, 
	HTStyle *, 		new,
	HTChildAnchor *,	info,		/* yuk */
	CONST char *,		saveName)	/* troppo do argumenti */
{
    int old_level, new_level, yca_level;
    nesting_state * old_nesting = (nesting_state*) old;
    nesting_state * new_nesting = (nesting_state*) new;
    nesting_state * yca;	/* youngest_common_ancestor */ ;
    
    {
        old_level = HTMLLevel(old);
        new_level = HTMLLevel(new);
	if(TRACE) {
		fprintf(stderr, "Was ");
		print_nesting(old_nesting);
		fprintf(stderr, ", will be ");
		print_nesting(new_nesting);
		fprintf(stderr, ".\n");
	}

	yca=new_nesting;
	
	/* Find equivalent point in new stack if deeper than old */
	for (yca_level=new_level; yca_level > old_level; yca_level--) {
	    yca = yca->parent;
	}

	/* Close levels to get to save level of nesting as new */
	for (; old_nesting != yca; old_level--) {
	    int t = old_nesting->element_number;
	    if (t>=0  	&& t!=HTML_HTML 
	    		&& DTD.tags[t].contents != SGML_EMPTY )
	        (*s->isa->end_element)(
			s, old_nesting->element_number); /* Close */
	    if (old_nesting->paragraph_break) {
	    	stripping = YES;
		SGML_gen_newlines = 0;
	    }
	    old_nesting = old_nesting->parent;
	    if (old_level <= yca_level) {
		yca = yca->parent;	/* move down together */
	    }
	}
	
	/* yca and old_nesting now point to Y.C.A */
#ifdef JUNK
	/* Close non-matching levels below that */
	while(old_nesting != yca) {
	    (*s->isa->end_element)(s, old_nesting->element_number); /* Close */
	    if (old_nesting->paragraph_break) {
	    	stripping = YES;
		SGML_gen_newlines = 0;
	    }
	    old_nesting = old_nesting->parent;
	    yca = yca->parent;
	}
#endif
	/* Open all levels back up to new level */
	while (yca != new_nesting){
	    nesting_state *p;
	    int t;
	    for(p=new_nesting; p->parent != yca; p=p->parent) /* loop */;
	    t = p->element_number;	/* Suppress outermost 2 levels */
	    if (!p->object_gens_HTML && t>=0  && t!=HTML_HTML) {
		if (t==HTML_A) {
		    start_anchor(s, info, saveName);
		} else {
		    (*s->isa->start_element)(s, t,
			    NULL, NULL); /* Open. @@@ No attributes! */
		    if (t==HTML_DL)		/* Kludge initial DT */
			(*s->isa->start_element)(s, HTML_DT, NULL, NULL);
		}
	    }
	    if (p->paragraph_break) {
	    	stripping = YES;
		SGML_gen_newlines = 0;
	    }
	    yca = p;
	}
    }
}


/*	Turn a text character into HTML
**	-------------------------------
**
** 	Within the prefix region (prefix!=0), we discard white space and
** 	characters matching *prefix++. Note the prefix string may contain
**	white space.
**
**	The SGML_gen_newlines flag means that newlines have been found.
**	They are not actually implemented unless some more non-white
**	text is found, so that
**	trailing newlines on the end of paragraphs are stripped.
**
*/
PUBLIC void HTRegenCharacter ARGS3(
	char,			c,
	HTStyle *,		style,
	HTStructured *,		target)
{
    nesting_state* 		nesting = (nesting_state*) style;
    white_space_treatment 	wst = nesting->wst;
/*    int 			element_number = nesting->element_number; */
    
#ifdef OLD_CODE
    if (prefix) {
	if (*prefix) {
	    if (c==*prefix) {
		++prefix;
		return;			/* Strip prefix characters */
	    }
	    if (WHITE(c)) return;			/* Strip white space */
	    if (TRACE) printf(
    "HTML: WARNING: Paragraph prefix incomplete: %i found where %i expected.\n",
			c, *prefix);
	    }
	prefix=0;				/* Prefix is over */
    }
#endif
    if (stripping) {
        if (WHITE(c)) return;
	stripping = NO;
    }
    
    if (nesting->element_number == HTML_LI) return;	/* Strip auto inserted */
    
    if (c=='\n') {				/* Paragraph Marks:	*/
	if (wst != WS_PRE) {			/* free format */
	    SGML_gen_newlines++;		/* Just flag it */
	    if (wst==WS_LIST) prefix = "\267\t";
	} else {
	    PUTC('\n');
	}
	
    } else {					/* Not newline */

	if (SGML_gen_newlines) {/* Got text, need paragraph break first */
	    if (c=='\t') {
	        if ((wst==WS_LIST) || (wst==WS_GLOSSARY)) {
	    	START(HTML_P);	/* Special case of P in list or glossary */
		return;
		}
	    }
	    switch(wst) {
	    case WS_NORMAL: 	START(SGML_gen_newlines>1? HTML_P : HTML_BR );
	    			break;
	    case WS_HEADING:	END(nesting->element_number);
	    			START(nesting->element_number);
	    			break;
	    case WS_DIR:
	    case WS_LIST:	START(HTML_LI); break;
	    case WS_GLOSSARY:	START(HTML_DT); break;
	    case WS_PRE:	PUTC('\n'); break;
	    }
	    SGML_gen_newlines=0;	/* paragraph flushed. */
	}
	
	/* Operator-inserted tags changes into SGML elements for some lists */
	if (c=='\t') {
	    switch(wst) {
		case WS_GLOSSARY:	if (nesting->element_number !=HTML_DD)
					    START(HTML_DD);
					break;
		case WS_DIR:		if (nesting->element_number !=HTML_LI)
					    START(HTML_LI); break;
		default:		PUTC('\t'); break;
	    }
	/* Not printable ASCII ==> check for entities */
	} else if ((c=='&') || (c=='<') ||
		((unsigned int)c > 127) || ((unsigned int)c < 32)) {
	    char **p = p_entity_values;
	    int i;
	    CONST SGML_dtd * dtd = &DTD;	/* @@ should pick up */
	    for(i=0; i<dtd->number_of_entities; i++, p++) {
	        if ( (**p == c) && ((*p)[1]==0)) {
		    (*target->isa->put_entity)(target, i);
		    return;
		}
	    }
	    if (TRACE) fprintf(stderr,
 "HTML: Can't find SGML entity with which to represent character %d ie `%c'\n",
 		(int)c, c);
	    PUTC(c);
	} else {				/* Not tab or entity */
	    PUTC(c);
	    
	} /* end not tab */
    } /* end not newline */

} /* HTRegenCharacter */


/*				BUILDING THE DOCUMENT
*/


/*		Make child nesting
**		------------------
**
**	If one exists, return that.
*/

PRIVATE nesting_state * make_child ARGS2(nesting_state *,parent,
		int, element_number)
{
    int i;
    HTList * list;
    nesting_state * child;
    
    if (!parent->children)
	parent->children = HTList_new();
    list = parent->children;
    for (i=0; i<HTList_count(list); i++) {
	child = (nesting_state*) HTList_objectAt(list,i);
	if (child->element_number == element_number)
	    return child;
    }

    child = (nesting_state*)calloc(1, sizeof(*child));
    if (child == NULL) outofmem(__FILE__, "nesting state");
    child->magic = NESTING_MAGIC_NUMBER;
    child->element_number = element_number;
    child->parent = parent;
    child->children = NULL;
    
    /* Copy style by default from parent */
    child->style = parent->style;
    if (child->style.paragraph.numTabs > 0) {
        int tab;
	child->style.paragraph.tabs = malloc(
		child->style.paragraph.numTabs *
			sizeof(child->style.paragraph.tabs[0]));
	for (tab=0; tab<child->style.paragraph.numTabs; tab++) {
	    child->style.paragraph.tabs[tab] = 
	    	parent->style.paragraph.tabs[tab];
	}
    }
    child->list_level = parent->list_level;
    child->wst = parent->wst;

    HTStyleModify(child, element_number);
    return child;
}


/*	Same but with HTStyle parameters
*/

PUBLIC HTStyle * HTNestElement ARGS2(HTStyle *, p, int, ele)
{
   return (HTStyle*) make_child((nesting_state*) p, ele);
}

/*	Ancestor
*/
PUBLIC HTStyle* HTMLAncestor ARGS2 (HTStyle *, old, int, depth)
{
   if (depth==0) return old;
   else return HTMLAncestor((HTStyle*)((nesting_state*)old)->parent, depth-1);
}


/*	Copy branch
*/
PUBLIC HTStyle* CopyBranch ARGS3 (HTStyle *, old, HTStyle *, new, int, depth)
{
   return depth ? CopyBranch (
   	    HTNestElement(old,
	 	((nesting_state*)HTMLAncestor(new, depth-1))->element_number),
	    new,
	    depth - 1)
	: old ;
}

/*	Add a slice in the tree
*/
PUBLIC HTStyle * HTInsertLevel ARGS3(HTStyle *, old,
		int,	element_number,
		int,	level)
{
    int how_far_back = HTMLLevel(old) - level;
    HTStyle * cut = HTMLAncestor(old, how_far_back);
    
    return CopyBranch(
    		HTNestElement(cut, element_number),
    		old,
		how_far_back);
}


/*	Cut a slice from the tree
*/
PUBLIC HTStyle * HTDeleteLevel ARGS2(HTStyle *, old,
		int,	level)
{
    int how_far_back = HTMLLevel(old) - level;
    HTStyle * cut = HTMLAncestor(old, how_far_back + 1 );
    
    return CopyBranch(
    		cut, 
    		old,
		how_far_back);
}

PUBLIC int HTMLElementNumber ARGS1(HTStyle *, s)
{
    return ((nesting_state*)s)->element_number; 
}

/*	Get a valid proper style for this document
**	which has nesting information
*/

PUBLIC HTStyle* HTNestingNamed ARGS2 (HTStructured*, me,
	CONST char *,	name)
{
    char * p = NULL;
    nesting_state * nesting = &me->nest;		/* base */
    StrAllocCopy(p, name);
    while (p){
	int t;
	HTTag* tag;
	char * q = strchr(p, '/');
	if (q) *q++ = 0;
	tag = SGMLFindTag(me->dtd, p);
    	if (!tag) return NULL;		/* Bad tag name */
	t = tag - &me->dtd->tags[0];	/* tag number */
	nesting = make_child(nesting, t);
	p = q;
    }
    return &nesting->style;
}


/* 	Return a string representation of a nesting state
**	-------------------------------------------------
*/
PUBLIC char * HTNestingName ARGS2(HTStructured*, me,
		HTStyle *, style)
{
    char * name = NULL;
    nesting_state * nesting = (nesting_state*) style;
    if (nesting->magic != NESTING_MAGIC_NUMBER) return NULL;
    
    for(;nesting; nesting = nesting->parent) {
        int t = nesting->element_number;
	if (t>=0) {
	    char * tn = me->dtd->tags[t].name;
	    if (name) {
	       char * s = malloc(strlen(name) + strlen(tn) + 2);
	       sprintf(s, "%s/%s", tn, name);
	       free(name);
	       name = s;
	    } else {
	        StrAllocCopy(name, tn);
	    }
	}
    }
    return name;
}


/*		Store a new element start
**		-------------------------
*/
PRIVATE void store_element_start ARGS2(HTStructured *, me, int, element_number)
{
/*  	Make or find a new nesting state element
*/	
    me->current_nesting = make_child(me->current_nesting, element_number);
    
    
/*	Make a new item in the document
**	-------------------------------
**
**	The child item is made in any case: if it is empty, it
**	is also closed.
*/
    if (me->store) {		/* Save new element */
	document_item * item;
	document_item * parent;

	item = (document_item*) calloc(1, sizeof(*item));
	if (item == NULL) outofmem(__FILE__, "document item");
	item->nesting = me->current_nesting;
	item->itype = HT_ELEMENT;
	
	for(parent = me->current_item; /* Find containing element */
		parent->itype != HT_DOCUMENT
		    && parent->itype != HT_ELEMENT;
		parent = parent->parent); /* void */
	item->parent = parent;
	if (!parent->as.element.contents) {
	    parent->as.element.contents = HTList_new();
	    HTList_addObject(parent->as.element.contents, item);
	}

	me->current_item = item;
    }
}


/*	Store element end
**	-----------------
*/

PRIVATE void store_element_end ARGS2(HTStructured *, me, int, element_number)
{

    if (me->store)
        me->current_item = me->current_item->parent; /* Pop state off stack */
    me->current_nesting = me->current_nesting->parent;
}  

/*_________________________________________________________________________
**
**	HTML-specific	M E T H O D S
*/


/*	Generate the style for a new combination
**	----------------------------------------
**
**	This SHOULD use the style sheet and allow incremental
**	changes for nesting things, and dependencies on attribute
**	values :-( .  Simple..
*/

#define LIST_INDENT 	30
#define GLOSSARY_INDENT 150

/*	Set a hangling indent and first tab stop
*/
PRIVATE void set_hanging_indent ARGS2(HTStyle *, style, int, i)
{
    NXTextStyle *p = &(style->paragraph);
    if (p->numTabs>0 && p->tabs) free(p->tabs);

    p->numTabs = 1;
    p->tabs = malloc(1 * sizeof(p->tabs[0]));
    p->tabs[0].kind = 0;
    p->tabs[0].x = p->indent1st + i;
    p->indent2nd = p->indent1st + i;
}


PRIVATE void HTStyleModify(nesting_state* nesting, int element_number)
{
    
    switch(element_number) {
    
        case HTML_HR: 		nesting->paragraph_break = YES;
				return;
	case HTML_IMG:
	case HTML_INPUT:	nesting->object_gens_HTML = YES;
				break;
	
        case HTML_A:            nesting->style.textRGBColor = 0x007700;
                                break;
	case HTML_B:
	case HTML_I:
	case HTML_STRONG:
	case HTML_CITE:
	case HTML_U:
	case HTML_EM:
	case HTML_TT:		return;
	
	case HTML_P:
	case HTML_BLOCKQUOTE:
	case HTML_ADDRESS:	nesting->wst = WS_NORMAL;
				nesting->paragraph_break = YES;
				break;
	case HTML_H1:
	case HTML_H2:
	case HTML_H3:
	case HTML_H4:
	case HTML_H5:
	case HTML_H6:		nesting->wst = WS_HEADING;
				nesting->paragraph_break = YES;
				break;
	case HTML_XMP:
	case HTML_LISTING:
	case HTML_PLAINTEXT:
	case HTML_PRE:		nesting->wst = WS_PRE;
				nesting->paragraph_break = YES;
				break;
	case HTML_UL:
	case HTML_OL:
	case HTML_MENU:		nesting->item_number = 0;
				nesting->style.paragraph.indent1st =
				    nesting->style.paragraph.indent2nd;

				set_hanging_indent(
				    &nesting->style,
				    LIST_INDENT);
				nesting->list_level++;

				nesting->wst = WS_LIST;
				nesting->paragraph_break = YES;
				return;  /* style is fixed up OK */
				
	case HTML_DL:
				nesting->item_number = 0;
				nesting->style.paragraph.indent1st =
				    nesting->style.paragraph.indent2nd;

				set_hanging_indent(
				    &nesting->style,
				    GLOSSARY_INDENT);
				nesting->list_level++;

				nesting->wst = WS_GLOSSARY;
				nesting->paragraph_break = YES;
				return;  /* style is fixed up OK */
				
	case HTML_DIR:		nesting->wst = WS_DIR;
				nesting->paragraph_break = YES;
				break;

	case HTML_FORM:		nesting->style.textGray = 0.667;
				nesting->style.textRGBColor = 0x770000;
				nesting->paragraph_break = YES;
				break;
				
	default:
				nesting->paragraph_break = NO;
				break;
    }  /* switch on element */
    
	/* @@@@@@@ placeholder only --- will go! @@@  */
    {
	HTStyle * s = styles[element_number];
	if (s) nesting->style = *s;
    }
}

/*	PUBLIC METHODS
*/

/*	Character handling
**	------------------
*/
PRIVATE void HTML_put_character ARGS2(HTStructured *, me, char, c)
{

    switch (me->current_nesting->element_number) {
    case HTML_COMMENT:
    	break;					/* Do Nothing */
	
    case HTML_TITLE:	
    	if (WHITE(c)) {
	   if (me->title.size ==0) return;	/* Strip leading WS */
	   c = ' ';
	}
	HTChunkPutc(&me->title, c);
	break;

	
    case HTML_LISTING:				/* Litteral text */
    case HTML_XMP:
    case HTML_PLAINTEXT:
    case HTML_PRE:
/*	We guarrantee that the style is up-to-date in begin_litteral
*/
    	HText_appendCharacter(me->text, c);
	break;
	
    default:					/* Free format text */
	if (me->style_change) {
	    if ((c=='\n') || (c==' ')) return;	/* Ignore it */
	    UPDATE_STYLE;
	}
	if (c=='\n' || c==' ') {	/* Collapse spaces too TBL 931106 */
	    if (me->in_word) {
		HText_appendCharacter(me->text, ' ');
		me->in_word = NO;
	    }
	} else {
	    HText_appendCharacter(me->text, c);
	    me->in_word = YES;
	}
    } /* end switch */
}



/*	String handling
**	---------------
**
**	This is written separately from put_character becuase the loop can
**	in some cases be promoted to a higher function call level for speed.
*/
PRIVATE void HTML_put_string ARGS2(HTStructured *, me, CONST char*, s)
{

    switch (me->current_nesting->element_number) {
    case HTML_COMMENT:
    	break;					/* Do Nothing */
	
    case HTML_TITLE:	
    	HTChunkPuts(&me->title, s);
	break;

	
    case HTML_LISTING:				/* Litteral text */
    case HTML_XMP:
    case HTML_PLAINTEXT:
    case HTML_PRE:

/*	We guarrantee that the style is up-to-date in begin_litteral
*/
    	HText_appendText(me->text, s);
	break;
	
    default:					/* Free format text */
        {
	    CONST char *p = s;
	    if (me->style_change) {
		for (; *p && ((*p=='\n') || (*p==' ')); p++)  ;  /* Ignore leaders */
		if (!*p) return;
		UPDATE_STYLE;
	    }
	    for(; *p; p++) {
		if (me->style_change) {
		    if ((*p=='\n') || (*p==' ')) continue;  /* Ignore it */
		    UPDATE_STYLE;
		}
		if (*p=='\n') {
		    if (me->in_word) {
			HText_appendCharacter(me->text, ' ');
			me->in_word = NO;
		    }
		} else {
		    HText_appendCharacter(me->text, *p);
		    me->in_word = YES;
		}
	    } /* for */
	}
    } /* end switch */
}




/*	Buffer write
**	------------
*/
PRIVATE void HTML_write ARGS3(HTStructured *, me, CONST char*, s, int, l)
{
    CONST char* p;
    CONST char* e = s+l;
    for (p=s; s<e; p++) HTML_put_character(me, *p);
}


/*	Start Element
**	-------------
*/
PRIVATE void HTML_start_element ARGS4(
	HTStructured *, 	me,
	int,		element_number,
	CONST BOOL*,	 	present,
	CONST char **,	value)
{

    store_element_start(me, element_number);
    change_paragraph_style(me, (HTStyle*)me->current_nesting);
    
    switch (element_number) {
    case HTML_A:			/* 	Anchors			*/
	{
	    HTChildAnchor * source;
	    char * href = NULL;
	    if (present[HTML_A_HREF]) {
	    	StrAllocCopy(href, value[HTML_A_HREF]);
		HTSimplify(href);
	    }
	    source = HTAnchor_findChildAndLink(
		me->node_anchor,				/* parent */
		present[HTML_A_NAME] ? value[HTML_A_NAME] : 0,	/* Tag */
		present[HTML_A_HREF] ? href : 0,		/* Addresss */
		present[HTML_A_REL] && value[HTML_A_REL] ? 
			(HTLinkType*)HTAtom_for(value[HTML_A_REL])
			    			: 0);
	    
	    if (present[HTML_A_TITLE] && value[HTML_A_TITLE]) {
	        HTParentAnchor * dest = 
		    HTAnchor_parent(
			HTAnchor_followMainLink((HTAnchor*)source)
				    );
		if (!HTAnchor_title(dest))
			HTAnchor_setTitle(dest, value[HTML_A_TITLE]);
	    }
	    UPDATE_STYLE;
	    HText_beginAnchor(me->text, source);
	}
    	break;

    case HTML_IMG:			/* Images -- insert them */
	{
	    HTChildAnchor * source;
	    char * href = NULL;
	    UPDATE_STYLE;
	    if (present[HTML_IMG_SRC]  && value[HTML_IMG_SRC]) {
	    	StrAllocCopy(href, value[HTML_IMG_SRC]);
		HTSimplify(href);
	    }
	    source = HTAnchor_findChildAndLink(
		me->node_anchor,				/* parent */
		0,						/* Id */
		present[HTML_IMG_SRC] ? href : 0,		/* Addresss */
		0);
	    
	    HText_appendImage(me->text, source,
	    	present[HTML_IMG_ALT] ? value[HTML_IMG_ALT] : 0,
		0 /* alignment*/,
		present[HTML_IMG_ISMAP]);
	}
	break;
	
    case HTML_INPUT:		/* Pass these on untouched */
        HText_appendObject(me->text, element_number, present, value);
	break;
	
    case HTML_HR:
	UPDATE_STYLE;
	HText_appendParagraph(me->text);
        HText_appendObject(me->text, element_number, present, value);
	HText_appendParagraph(me->text);
        break;
	
    case HTML_TITLE:
        HTChunkClear(&me->title);
	break;
	
    case HTML_NEXTID:
    	/* if (present[NEXTID_N] && value[NEXTID_N])
		HText_setNextId(me->text, atoi(value[NEXTID_N])); */
    	break;
	
    case HTML_ISINDEX:
   	HTAnchor_setIndex(me->node_anchor);
	break;
	
    case HTML_P:
	UPDATE_STYLE;
	HText_appendParagraph(me->text);
	if ((me->current_nesting->parent->wst == WS_LIST) 
	     || (me->current_nesting->parent->wst == WS_GLOSSARY))
	    HText_appendCharacter(me->text, '\t');
	me->in_word = NO;
	break;

    case HTML_DL:				/* Glossaries */
    
        /* change_paragraph_style(me, present && present[DL_COMPACT]
    		? styles[HTML_DL]
		: styles[HTML_DL]); */
	me->para_break_on_new_style = YES;
	break;
	
    case HTML_DT:
        if (!me->style_change) {
	   /*  HText_appendParagraph(me->text); */
	    HText_appendCharacter(me->text, '\n');
	    me->in_word = NO;
	}
	me->para_break_on_new_style = YES;
	UPDATE_STYLE;
	break;
	
    case HTML_DD:
        UPDATE_STYLE;
	HText_appendCharacter(me->text, '\t');
	me->in_word = NO;
	break;

    case HTML_UL:
    case HTML_OL:
    case HTML_MENU:
    case HTML_DIR:
	/* change_paragraph_style(me, styles[element_number]); */
	me->para_break_on_new_style = YES;
	break;
	
    case HTML_LI:
        {
	    BOOL para_done = me->style_change && me->para_break_on_new_style;
	    UPDATE_STYLE;
	    if (me->current_nesting->parent->element_number != HTML_DIR) {
		int i = me->current_nesting->parent->item_number++;
		int level = me->current_nesting->parent->list_level;
		if (!para_done) HText_appendParagraph(me->text);
		if (me->current_nesting->parent->element_number == HTML_OL) {
		    char junk[10];
		    sprintf(junk, "%d)", i+1);
		    HText_appendText(me->text, junk);
		} else {
		    HText_appendCharacter(me->text, bullets[level-1]);
		}
		HText_appendCharacter(me->text, '\t');
	    } else {
		HText_appendCharacter(me->text, '\t');	/* Tab @@ nl for UL? */
	    }
	    me->in_word = NO;
	    break;
	}
    case HTML_LISTING:				/* Litteral text */
    case HTML_XMP:
    case HTML_PLAINTEXT:
    case HTML_PRE:
	me->para_break_on_new_style = YES;

	UPDATE_STYLE;
    	if (me->comment_end)
    	    HText_appendText(me->text, me->comment_end);
	break;

	
	          
    case HTML_HTML:			/* Ignore these altogether */
    case HTML_HEAD:
    case HTML_BODY:
    	break;
	
    case HTML_TT:			/* Physical character highlighting */
    case HTML_B:			/* Currently ignored */
    case HTML_I:
    case HTML_U:
    
    case HTML_EM:			/* Logical character highlighting */
    case HTML_STRONG:			/* Currently ignored */
    case HTML_CODE:
    case HTML_SAMP:
    case HTML_KBD:
    case HTML_VAR:
    case HTML_DFN:
    case HTML_CITE:
    	break;
	
    case HTML_H1:			/* paragraph styles */
    case HTML_H2:
    case HTML_H3:
    case HTML_H4:
    case HTML_H5:
    case HTML_H6:
    case HTML_H7:
    case HTML_ADDRESS:
    case HTML_BLOCKQUOTE:
	me->para_break_on_new_style = YES;
	break;

    } /* end switch */

    if (me->dtd->tags[element_number].contents == SGML_EMPTY) {
    	store_element_end(me, element_number);
	change_paragraph_style(me, (HTStyle*)me->current_nesting);
    }
}


/*		End Element
**		-----------
**
*/
/*	When we end an element, the style must be returned to that
**	in effect before that element.  Note that anchors (etc?)
**	don't have an associated style, so that we must scan down the
**	stack for an element with a defined style. (In fact, the styles
**	should be linked to the whole stack not just the top one.)
**	TBL 921119
**
**	We don't turn on "CAREFUL" check because the parser produces
**	(internal code errors apart) good nesting. The parser checks
**	incoming code errors, not this module.
*/
PRIVATE void HTML_end_element ARGS2(HTStructured *, me, int , element_number)
{
#ifdef CAREFUL		/* parser assumed to produce good nesting */
    if (element_number != me->current_nesting->element_number) {
        fprintf(stderr,
	    "HTMLText: end of element %s when expecting end of %s\n",
	    me->dtd->tags[element_number].name,
	    me->dtd->tags[me->current_nesting->element_number].name);
	/* panic */
    }
#endif
    
    if (me->store)
        me->current_item = me->current_item->parent; /* Pop state off stack */
    me->current_nesting = me->current_nesting->parent;
    
    switch(element_number) {

    case HTML_A:
	change_paragraph_style(me, &me->current_nesting->style);
	HText_endAnchor(me->text);
	UPDATE_STYLE;
	break;

    case HTML_TITLE:
        HTChunkTerminate(&me->title);
    	HTAnchor_setTitle(me->node_anchor, me->title.data);
	break;
	
    case HTML_HTML:			/* Ignore these altogether */
    case HTML_HEAD:
    case HTML_BODY:
    	break;
	
    case HTML_TT:			/* Physical character highlighting */
    case HTML_B:			/* Currently ignored */
    case HTML_I:
    case HTML_U:
    
    case HTML_EM:			/* Logical character highlighting */
    case HTML_STRONG:			/* Currently ignored */
    case HTML_CODE:
    case HTML_SAMP:
    case HTML_KBD:
    case HTML_VAR:
    case HTML_DFN:
    case HTML_CITE:
	change_paragraph_style(me, &me->current_nesting->style);
    	break;

    case HTML_LISTING:				/* Litteral text */
    case HTML_XMP:
    case HTML_PLAINTEXT:
    case HTML_PRE:
    	if (me->comment_start)
    	    HText_appendText(me->text, me->comment_start);
	/* Fall through */
	
    case HTML_H1:			/* paragraph styles */
    case HTML_H2:
    case HTML_H3:
    case HTML_H4:
    case HTML_H5:
    case HTML_H6:
    case HTML_H7:
    case HTML_ADDRESS:
    case HTML_BLOCKQUOTE:
    
    case HTML_DL:
    case HTML_UL:
    case HTML_OL:
    case HTML_MENU:
    case HTML_DIR:

	change_paragraph_style(me, &me->current_nesting->style);
	me->para_break_on_new_style = YES;
		/* Often won't really change */
	break;
	
    } /* switch */
}


/*		Expanding entities
**		------------------
*/
/*	(In fact, they all shrink!)
*/

PRIVATE void HTML_put_entity ARGS2(HTStructured *, me, int, entity_number)
{
    HTML_put_string(me, p_entity_values[entity_number]); /* @@ Other representations */
}

/*	Free a nesting state's descendents
*/

PRIVATE void free_nesting ARGS1(nesting_state *, p)
{
    nesting_state * q;
    if (p->children) {
	while(( q = HTList_removeFirstObject(p->children)) != NULL) {
	    free_nesting(q);
	    free(q);
	}
	HTList_free(p->children);
    }
}

/*	Free an HTML object
**	-------------------
**
** If the document is empty, the text object will not yet exist.
   So we could in fact abandon creating the document and return
   an error code.  In fact an empty document is an important type
   of document, so we don't.
**
**	If non-interactive, everything is freed off.   No: crashes -listrefs
**	Otherwise, the interactive object is left.	
*/
PUBLIC void HTML_free ARGS1(HTStructured *, me)
{
    if (!me->build_finished) {
	    UPDATE_STYLE;		/* Creates empty document here! */
	if (me->comment_end)
		    HTML_put_string(me,me->comment_end);
	HText_endAppend(me->text);
 
	if (me->target) {
	    (*me->targetClass.free)(me->target);
	}
	if (me->request->childAnchor) {
	    HText_selectAnchor(me->text, me->request->childAnchor);
	}
	me->build_finished = YES;
    }
    if (--me->reference_count == 0) {
    	free_nesting(&me->nest);
    	free(me);
    }
}


PUBLIC void HTML_reopen ARGS1(HTStructured *, me)
{
    me->reference_count++;
}

/* Error close */

PRIVATE void HTML_abort ARGS2(HTStructured *, me, HTError, e)

{
    if (!me->build_finished) {
	if (me->target) {
	    (*me->targetClass.abort)(me->target, NULL);
	}
	me->build_finished = YES;
    }
    if (--me->reference_count == 0) {
    	free(me);
    }
}


/*	Get Styles from style sheet
**	---------------------------
*/
PRIVATE void get_styles NOARGS
{
    got_styles = YES;
    
    default_style =		HTStyleNamed(styleSheet, "Normal");

    styles[HTML_H1] =		HTStyleNamed(styleSheet, "Heading1");
    styles[HTML_H2] =		HTStyleNamed(styleSheet, "Heading2");
    styles[HTML_H3] =		HTStyleNamed(styleSheet, "Heading3");
    styles[HTML_H4] =		HTStyleNamed(styleSheet, "Heading4");
    styles[HTML_H5] =		HTStyleNamed(styleSheet, "Heading5");
    styles[HTML_H6] =		HTStyleNamed(styleSheet, "Heading6");
    styles[HTML_H7] =		HTStyleNamed(styleSheet, "Heading7");

    styles[HTML_DL] =		HTStyleNamed(styleSheet, "Glossary");
    styles[HTML_UL] =
    styles[HTML_OL] =		HTStyleNamed(styleSheet, "List");
    styles[HTML_MENU] =		HTStyleNamed(styleSheet, "Menu");
    styles[HTML_DIR] =		HTStyleNamed(styleSheet, "Dir");    
/*  styles[HTML_DLC] =		HTStyleNamed(styleSheet, "GlossaryCompact"); */
    styles[HTML_ADDRESS]=	HTStyleNamed(styleSheet, "Address");
    styles[HTML_BLOCKQUOTE]=	HTStyleNamed(styleSheet, "BlockQuote");
    styles[HTML_PLAINTEXT] =
    styles[HTML_XMP] =		HTStyleNamed(styleSheet, "Example");
    styles[HTML_PRE] =		HTStyleNamed(styleSheet, "Example");
    styles[HTML_LISTING] =	HTStyleNamed(styleSheet, "Listing");
}
/*				P U B L I C
*/

/*	Structured Object Class
**	-----------------------
*/
PUBLIC CONST HTStructuredClass HTMLPresentation = /* As opposed to print etc */
{		
	"text/html",
	HTML_free,
	HTML_abort,
	HTML_put_character, 	HTML_put_string,  HTML_write,
	HTML_start_element, 	HTML_end_element,
	HTML_put_entity
}; 


/*		New Structured Text object
**		--------------------------
**
**	The strutcured stream can generate either presentation,
**	or plain text, or HTML.
*/
PUBLIC HTStructured* HTML_new ARGS5(
	HTRequest *,		request,
	void *,			param,
	HTFormat,		input_format,
	HTFormat,		output_format,
	HTStream *,		output_stream)
{

    HTStructured * me;
    
    if (output_format != WWW_PLAINTEXT
    	&& output_format != WWW_PRESENT
	&& output_format != HTAtom_for("text/x-c")) {
        HTStream * intermediate = HTStreamStack(WWW_HTML, request, YES);
	if (intermediate) return HTMLGenerator(intermediate);
        fprintf(stderr, "** Internal error: can't parse HTML to %s\n",
       		HTAtom_name(output_format));
	exit (-99);
    }

    me = (HTStructured*) calloc(1, sizeof(*me));
    if (me == NULL) outofmem(__FILE__, "HTML_new");

    if (!got_styles) get_styles();

    me->isa = &HTMLPresentation;
    me->node_anchor =  request->anchor;
    me->title.size = 0;
    me->title.growby = 128;
    me->title.allocated = 0;
    me->title.data = 0;
    me->text = 0;
    me->style_change = YES; /* Force check leading to text creation */
    me->new_style = default_style;
    me->para_break_on_new_style = NO;
    me->old_style = 0;
    me->dtd = &DTD;	/* Personalise to HTML */
    me->comment_start = NULL;
    me->comment_end = NULL;
    me->target = output_stream;
    if (output_stream) me->targetClass = *output_stream->isa;	/* Copy pointers */
    
    
    /* Set up the root of the nesting state tree */
    
    me->nest.style = *HTStyleNamed(styleSheet, "Normal");	/* @@@ */
    me->nest.paragraph_break = YES;
    me->nest.parent	= NULL;
    me->nest.element_number = -1;
    me->nest.children = NULL;
    me->current_nesting = & me->nest;

    /*  Set up the root of the document tree */
    
    me->document.itype = HT_DOCUMENT;
    me->document.parent = NULL;
    me->document.nesting = & me->nest;
    me->document.total_chars = 0;
    me->document.total_lines = 0;
    me->document.as.element.contents = NULL;
    me->current_item = & me->document;
    me->store = NO;		/* Don't update this tree */
    me->reference_count = 1;	/* Deleet after one free */
    me->request = request;
    
    return (HTStructured*) me;
}


/*	HTConverter for HTML to plain text
**	----------------------------------
**
**	This will convert from HTML to presentation or plain text.
*/
PUBLIC HTStream* HTMLToPlain ARGS5(
	HTRequest *,		request,
	void *,			param,
	HTFormat,		input_format,
	HTFormat,		output_format,
	HTStream *,		output_stream)
{
    return SGML_new(&DTD, HTML_new(request, param,
    	input_format, output_format, output_stream));
}


/*	HTConverter for HTML to C code
**	------------------------------
**
**	C copde is like plain text but all non-preformatted code
**	is commented out.
**	This will convert from HTML to presentation or plain text.
*/
PUBLIC HTStream* HTMLToC ARGS5(
	HTRequest *,		request,
	void *,			param,
	HTFormat,		input_format,
	HTFormat,		output_format,
	HTStream *,		output_stream)
{
    
    HTStructured * html;
    
    (*output_stream->isa->put_string)(output_stream, "/* ");
    	/* Before even title */
    html = HTML_new(request, param, WWW_PLAINTEXT,
    		output_format, output_stream);
    html->comment_start = "/* ";
    html->comment_end = " */\n";	/* Must start in col 1 for cpp */
/*    HTML_put_string(html,html->comment_start); */
    return SGML_new(&DTD, html);
}


/*	Presenter for HTML
**	------------------
**
**	This will convert from HTML to presentation or plain text.
**
**	Override this if you have a windows version
*/
#ifndef GUI
PUBLIC HTStream* HTMLPresent ARGS5(
	HTRequest *,		request,
	void *,			param,
	HTFormat,		input_format,
	HTFormat,		output_format,
	HTStream *,		output_stream)
{
    return SGML_new(&DTD, HTML_new(request,
    		param, input_format,
    		output_format, output_stream));
}
#endif


/*	Record error message as a hypertext object
**	------------------------------------------
**
**	The error message should be marked as an error so that
**	it can be reloaded later.
**	This implementation just throws up an error message
**	and leaves the document unloaded.
**	A smarter implementation would load an error document,
**	marking at such so that it is retried on reload.
**
** On entry,
**	sink 	is a stream to the output device if any
**	number	is the HTTP error number
**	message	is the human readable message.
**
** On exit,
**	returns	a negative number to indicate lack of success in the load.
*/

PUBLIC int HTLoadError ARGS3(
	HTRequest *, 	request,
	int,		number,
	CONST char *,	message)
{
    HTAlert(message);		/* @@@@@@@@@@@@@@@@@@@ */
    return -number;
} 

