/*	Style Definition for Hypertext				HTStyle.h
**	==============================
**
**	Styles allow the translation between a logical property of a piece of
**	text and its physical representation.
**
**	A StyleSheet is a collection of styles, defining the
**	translation necessary to represent a document.
**	It is a linked list of styles.
*/

#import <appkit/appkit.h>
#import "libwww.h"

/* #include "HTUtils.h" */

#ifndef HTSTYLE_H
#define HTSTYLE_H

#define STYLE_NAME_LENGTH	80

typedef enum _SGML_tagtype {
	NONE,			/* Style holds until further notice 	*/
	ENDTAG, 		/* Style holds until end tag </xxx> 	*/
	LINE 			/* Style holds until end of line (ugh!)	*/
} SGML_tagtype;
	 
typedef NXCoord HTCoord;

/*	This HTStyle type is a subclass of NXTextStyle
**	so that the HTStyle* pointer can be used as a paragraph style
**	and the HTStyle* can be deduced back from the NXTextStyle*
*/

/* typedef struct _HTStyle HTStyle; */
struct _HTStyle {
	NXTextStyle	paragraph;	/* 0 lineHt means not defined */
	
	int		magic;		/* Convince us that the rest follows*/
#define HTSTYLE_MAGIC_NUMBER 50414694	/* Put this in and it is valid */

	struct _HTStyle	*next;		/* Link for putting into stylesheet */
	char *		name;		/* Style name */
	char *		SGMLTag;	/* Tag name to start */
	SGML_tagtype	SGMLType;	/* How to end it */

	id		font;		/* The character representation */
	HTCoord		fontSize;	/* The size of font, not independent */
	unsigned int	underline;	/* See Text.h's rFlags.underline */
#ifdef V1
	float		textColor;	/* Colour of text */
#else
	float		textGray;	/* Gray level of text */
	int		textRGBColor;	/* Colour levels of text */
#endif
	HTCoord		spaceBefore;	/* Omissions from NXTextStyle */
	HTCoord		spaceAfter;
	void		*anchor;	/* Anchor id if any, else zero ?? */
	
	HTNesting	*nesting;	/* SGML nesting state */
};

extern HTStyle * HTStyleForRun(HTStyleSheet *self, NXRun *run);
extern HTStyle * HTStyleForParagraph(HTStyleSheet * self, NXTextStyle *para);


#endif /* HTNSSTYLE_H */

