//  HyperCell.m
//
//  a simple cell subclass that can be placed into a text object
//
//  (c) TBL CERN 1993
//
#import "HyperCell.h"

#define THEDIRECTIVE "WWWHyperCell"

@implementation HyperCell

//#import "HTML.h"
//#import "HTFWriter.h"		/* For save and call back */
//#import "HTFile.h"		/* For file suffix lookup */
//#import "HyperText.h"		/* for getting nodeAnchor */

struct _HTStructured {		/* All we need to know */
	HTStructuredClass * isa;
	/* ... */
};

//sent when the class receives its first message
+ initialize
{
	[Text registerDirective:THEDIRECTIVE forClass:[HyperCell class]];
	return self;
}

+ new
{
    return [super new];
}

- init
{
    self = [super init];
    linkAnchor = NULL;
    present = NULL;
    value = NULL;
    return self;
}


//	Create object from Instance Variables
//	-------------------------------------
	
- setup
{
    switch(element_number){
    case HTML_HR:
	[self initIconCell:"hrule_fade_8gray"];
	break;
	
    case HTML_INPUT:		/** Much more complex! */
	[self initTextCell: present && present[HTML_INPUT_VALUE] ?
			value[HTML_INPUT_VALUE] : "(no value)" ];
	[self setEditable:YES];
	[self setBezeled:YES];
	[self setEnabled: !(present && present[HTML_INPUT_DISABLED])];
	
	if (present && present[HTML_INPUT_TYPE] && value[HTML_INPUT_TYPE]) {
		/* Types are: TEXT, INT, FLOAT, DATE, URL, CHECKBOX, RADIO,
		RANGE, IMAGE, SCRIBLE, AUDIO, SUBMIT, RESET */
		
	    char * v = value[HTML_INPUT_TYPE];
	    if (!strcasecomp(v, "TEXT")) [self setEntryType: NX_ANYTYPE];
	    else if (!strcasecomp(v, "INT")) [self setEntryType: NX_INTTYPE];
	    else if (!strcasecomp(v, "FLOAT"))
	    	[self setEntryType: NX_FLOATTYPE];
	    else if (!strcasecomp(v, "DATE")) [self setEntryType: NX_ANYTYPE];
	    else if (!strcasecomp(v, "URL")) [self setEntryType: NX_ANYTYPE];
	    else if (!strcasecomp(v, "CHECKBOX"))
	    	[self setEntryType: NX_ANYTYPE];
	}
	
	break;
	
    default:
        [self initIconCell:"unknown_24"];
	break;
    }
    return self;
}


//	Apparatus to build the cell from the web
//	----------------------------------------


- initForElement: (int)ele
	inDTD: 			(CONST SGML_dtd*) 	d
	attributesPresent:	(BOOL*)		pres
	withValues:		(char**)	val
	inView: 		(HyperText *)	v
{
   int a;
   int attributes = d->tags[ele].number_of_attributes;

//   [self init];
   /* Copy over SGML information */
   element_number = ele;
   dtd = d;
   view = v;
   present = NULL;
   value = NULL;
   if (attributes && pres) {
	present = (BOOL*)calloc(attributes, sizeof(BOOL));
	value = (char**) calloc(attributes, sizeof(char*));
	for (a=0; a<attributes; a++) {
	    present[a] = pres[a];
	    if (present[a]) StrAllocCopy(value[a],val[a]);
	}
    }
    [self setup];
    return self;
}

- free
{
   int a;
   int attributes = dtd->tags[element_number].number_of_attributes;
    if (present) {
        for(a=0; a<attributes; a++)
	    if (present[a] && value[a]) free(value[a]);
	free(present);
    }
    if (value) free(value);
    return [super free];
}


- (HTChildAnchor*) linkAnchor  		{ return linkAnchor; }

- setLinkAnchor: (HTChildAnchor*) anc
{
    linkAnchor = anc;
    [self setBordered: (anc != 0 )]; /* neat */
    [self setBezeled: anc && (HTAnchor_followMainLink((HTAnchor*) anc) != 0 )]; /* neat */
    return self;
}


//		Override Cell Methods
//
//	The Cell object calls itself with this just before drawing
//	or editing a cell's text. We can define the behaviour of this
//	by overriding these.  Note that the brackground grey is not drawn!
//
#ifdef NOPE
 - setTextAttributes:textObject
{
    [super setTextAttributes:textObject];	// See Cell documentation!
    [textObject setBackgroundGray: 0.3333];	
    return TextObject;
}
#endif

//	Read Rich Text
//	--------------
//
//	These routines allow cut and paste of objects.
//	The default rich text format is a mapping of the SGML format.

- readRichText:(NXStream *)stream forView:v
{
	char buffer[FILENAME_MAX+1];
	char buffer2[FILENAME_MAX+1];
	HTTag * t;
	int	a;
	
	view = v;
	dtd = &HTMLP_dtd;	// @@@@ Bug: Should pick up from view @@@@@
	
	NXScanf(stream,"%s %s\n",buffer, buffer2);
	t = SGMLFindTag(dtd, buffer);
	
	if (!t){
	    fprintf(stderr, "HyperCell: RTF element %s is not in DTD!\n",
	    	buffer);
	    return nil;
	}
	
	element_number = t - dtd->tags;

	if (strcmp(buffer2, "{") != 0) {
	    fprintf(stderr, "HyperCell: RTF element has no ""}""\n");
	    return nil;
	}

	for(a=0;a < t->number_of_attributes; a++) {
				/* Limit in case garbage RTF */
	    int i;
	    char * p;
	    int fields = NXScanf(stream,"%[^\n]\n",buffer);
	    if (fields==0) break;
	    p = strchr(buffer, ' ');
	    if (strcmp(buffer, "}")==0) break;
	    if (p) *p++ = 0;	/* split */
	    i = SGMLFindAttribute(t,buffer);
	    if (i<0) {
		fprintf(stderr, "HyperCell: Bad attribute name `%s'\n",
			buffer);
		return nil;
	    }
	    if (!present) {
		present = (BOOL*)calloc(t->number_of_attributes, sizeof(BOOL));
		value = (char**) calloc(t->number_of_attributes,
						sizeof(char*));
	    }
	    present[i] = YES;
	    if (p) {
	        value[i] = (char*)malloc(strlen(p)+1);
		strcpy(value[i], p);
	    }
	}
	
	
	[self setup];
			
//	[v replaceSelWithCell:self];

	return self;
}


//		Write out Rich Text
//		-------------------
//

- writeRichText:(NXStream *)stream forView:view
{
    int attributes = dtd->tags[element_number].number_of_attributes;
    int a;
    
    NXPrintf(stream,"%s {\n", dtd->tags[element_number].name);
    for(a=0; a<attributes; a++) {
	if (present && present[a]) {
	    if (value[a]) {
		NXPrintf(stream,"%s %s\n",
		    dtd->tags[element_number].attributes[a].name,
		    value[a]);	/* Bug: Fails with emtpty string values */
	    } else {
		NXPrintf(stream,"%s\n",
		    dtd->tags[element_number].attributes[a].name);
	    }
	}
    }
    NXPrintf(stream,"}\n");

    return self;
}

//		Write SGML back
//		---------------

- writeSGML:(HTStructured *) thisTarget relativeTo:(const char *)aName
{
    
    (*thisTarget->isa->start_element)(thisTarget,
    		element_number, present, value);

    return self;
}

- (BOOL)trackMouse:(NXEvent *)theEvent inRect:(const NXRect *)rect ofView:view
{
	return NO;
}

@end
