//	HyperText Access method manager Object			HyperManager.m
//	--------------------------------------
//
//
// History:
//	   Oct 90	Written TBL
//
//	(c) TBL,CERN
//
#define NextStep	/* Control HText.h */

#define HELP_ENTRY "http://info.cern.ch/hypertext/WWW/NextStep/WorldWideWeb.html"

#import "HyperManager.h"

#import "HyperText.h"

#import "libwww.h"

@implementation HyperManager 

#import "WWWPageLayout.h"
#import <appkit/appkit.h>

#define THIS_TEXT  (HyperText *)[[[NXApp mainWindow] contentView] docView]

extern char * WWW_nameOfFile(const char * name);	/* In file access */

/*	Exported to everyone */

int WWW_TraceFlag;	/* Exported to everyone */
char * appDirectory;	/* Name of the directory containing the application */

PUBLIC HyperAccess* TheManager;
PUBLIC HTRequest *  PresentRequest;	/* Used by HyperText */

+ new
{
    self = [super new];
    accesses = [List new];		// Create and clear list
    TheManager = self;			// For text objects to find 
    request = HTRequest_new();		// Use this for all normal requests
    HTFormatInit(request->conversions); // Set up the conversions we handle.
    PresentRequest = request;
    
    [self revertPreferences:nil];
    
    return self;
}

- traceOn:sender { WWW_TraceFlag = 1; return self;}
- traceOff:sender { WWW_TraceFlag = 0; return self;}
- (HTRequest*) request { return request;}

- (const char *) name
{
    return "any";
}

/////////////////////////////////////////////////////////////////////
//
//			Preferences

extern BOOL HTSuppressImages;
PUBLIC BOOL HTCacheHTTP = NO;
extern BOOL HTMultiWindowMode;
PUBLIC char * MyIconRoot = NULL;

//	Update the internals of the program from the preferences
//	panel.  Should be called whenever the panel changes.

- updatePreferences:sender
{

    CONST char * s;
    
    s = [preferenceStrings stringValueAt:2];
    if (s && *s) {
        StrAllocCopy(MyIconRoot,s);
	HTStdIconInit(s);
    }
    
    HTSuppressImages =	[[preferenceSwitches cellAt:0:0] intValue] !=0 ;
    HTCacheHTTP = 	[[preferenceSwitches cellAt:1:0] intValue] !=0 ;
    HTMultiWindowMode = [[preferenceSwitches cellAt:2:0] intValue] !=0 ;
    

    WWW_TraceFlag =	[[debugSwitches cellAt:0:0] intValue] !=0 ;
    return self;
}

//	Save preferences panel to defaults
//
- savePreferences:sender
{
    CONST char *s;
    
    [self updatePreferences: nil];


    s = [preferenceStrings stringValueAt:0];
    NXWriteDefault("WorldWideWeb", "NewsHost",s);
	
    s = [preferenceStrings stringValueAt:1];
    NXWriteDefault("WorldWideWeb", "Proxy",s);

    s = [preferenceStrings stringValueAt:2];
    NXWriteDefault("WorldWideWeb", "IconRoot",s);

    NXWriteDefault("WorldWideWeb", "SuppressImages",
    	HTSuppressImages ? "YES" : "NO");
    NXWriteDefault("WorldWideWeb", "CacheHTTP",
    	HTCacheHTTP ? "YES" : "NO");
    NXWriteDefault("WorldWideWeb", "MultiWindow",
    	HTMultiWindowMode ? "YES" : "NO");

//	Debug

    NXWriteDefault("WorldWideWeb", "MallocDebug",
    	[debugStrings stringValueAt:0]);

    NXWriteDefault("WorldWideWeb", "Trace",
    	[[debugSwitches cellAt:0:0] intValue] ? "YES" : "NO");
    return self;
}


//	Load preferences: Note if strings YES and NO not given
//	defaults may vary but try to keep them absent = No = off.

- revertPreferences:sender
{
    CONST char * s;
    [preferenceStrings setStringValue:
    	NXReadDefault("WorldWideWeb", "NewsHost") at:0];
	
    [preferenceStrings setStringValue:
    	NXReadDefault("WorldWideWeb", "Proxy") at:1];
	
    s = NXReadDefault("WorldWideWeb", "SuppressImages");
    [[preferenceSwitches cellAt:0:0] setIntValue:
    	s? (strcasecomp(s, "YES")==0) :NO];
    
    s = NXReadDefault("WorldWideWeb", "CacheHTTP");
    [[preferenceSwitches cellAt:1:0] setIntValue:
    	s? (strcasecomp(s, "YES") == 0): NO];
    
    s = NXReadDefault("WorldWideWeb","MultiWindow");
    [[preferenceSwitches cellAt:2:0] setIntValue:
    	s? (strcasecomp(s,"YES")==0): NO];

//	Debug bits

    [debugStrings setStringValue:
    	NXReadDefault("WorldWideWeb", "MallocDebug") at:0];

    s = NXReadDefault("WorldWideWeb","Trace");
    [[debugSwitches cellAt:0:0] setIntValue:
    	s ? (strcasecomp(s,"YES")==0): NO];
    
    return [self updatePreferences:self];
	 
}

//			Access Management functions
//


//	Load an anchor				loadAnchor:
//	--------------
//
//	It checks whether in fact the anchor
//	is already loaded and linked, and that the address string is not null.
//
// On exit:
//	If a duplicate node is found, that anchor is returned
//	If there is no success, nil is returned.
//	Otherwise, the anchor is returned.

- loadAnchor:(HTAnchor *)anAnchor Diagnostic:(int)diagnostic
{

    HText * HT;
    HTRequest_clear(request);
    if (!HTLoadAnchor(anAnchor, request)) return nil;
    HTHistory_record(anAnchor);
    //	The node may have become an index: update the existence
    //   state of the panel.
    
    HT = (HText*) HTAnchor_document(HTAnchor_parent(anAnchor));
    if (HTAnchor_isIndex([HT nodeAnchor])) {
	[[keywords window] makeKeyAndOrderFront:self];
    } else {
	[[keywords window] close];
//		[[keywords window] orderOut:self];    @@ bug?
    }
    return self;
}

//______________________________________________________________________________


//	Open or search  by name
//	-----------------------
//
//	@@@ diagnostcics lost
//
- accessName:(const char *)arg
	Diagnostic:(int)diagnostic
{
    int status;
    if (!arg || !*arg) return nil;
    HTRequest_clear(request);
    status = HTLoadAbsolute(arg, request);
    return status ? self : nil;
}


//	Search with a given diagnostic level
//
//	This involves making a special address string, being the index address
//	with a ? sign followed by a "+" separated list of keywords.
//
- searchDiagnostic:(int)diagnostic
{

    HyperText * HT = THIS_TEXT;
    if (!HT) return nil;
    
    HTRequest_clear(request);
    HTSearch([keywords stringValueAt:0], [HT nodeAnchor], request);
    HTHistory_record((HTAnchor*)HTMainAnchor);
    return self;
}

//				N A V I G A T I O N


//	Relative moves
//	--------------
//
//	These navigate around the web as though it were a tree, from the point of
//	view of the user's browsing order.

- back:sender
{ 
    HTRequest_clear(request);
    HTLoadAnchor(HTHistory_backtrack(), request);
    return self;
}

- next:sender
{
    HTRequest_clear(request);
    HTLoadAnchor(HTHistory_next, request);
    return self;
}

- previous:sender
{ 
    HTRequest_clear(request);
    HTLoadAnchor(HTHistory_previous, request);
    return self;
}


//	Load a personal or system-wide version of a file
//	------------------------------------------------
//
//	This accesses the personal copy for the user or, failing that,
//	for the system. It is important that this name is fully qualified
//	as other names will be generated relative to it.
//	We check whether the local version is there before loading it
//	in order to prevent error messages being given to the user if
//	it is not.
//
// On exit,
//	Returns	If successful, the anchor of the node loaded or found.
//		If failed, nil.

- (HTParentAnchor*) openMy:(const char *)filename diagnostic:(int)diagnostic
{
    HTParentAnchor * myAnchor;
    HTParentAnchor * templateAnchor;
    char name[256];
    char template[256];
    
    if (getenv("HOME")) {
	int status;
	struct stat buf;			/* status buffer */
	strcpy(name, getenv("HOME"));
	strcat(name, "/WWW");
        status = stat(name, &buf);		/* Does dir exist? */
	if (status<0) {   /* no */
	    status = mkdir(name, 0777);		/* all rwx */
	    if (TRACE) fprintf(stderr, "Creation of %s returned %s\n",
	    		name, status ? "** ERROR **" : "success" );
	}
	strcat(name, "/");
	strcat(name, filename);
        status = stat(name, &buf);		/* Does file exist? */
	{
	    char * myName = WWW_nameOfFile(name);
	    myAnchor = (HTParentAnchor*)HTAnchor_findAddress(myName);
	    free(myName);
        }
	if (TRACE) fprintf(stderr,"File: stat(%s) returned %d\n",
				name, status);
	if (status>=0) {			/* If it does, load it */
	    HTRequest_clear(request);
	    HTLoadAnchor((HTAnchor*)myAnchor, request);
	    return myAnchor;
        }
    } else {
    	return NULL;
    }


   /* There is no local copy yet
   */
    strcpy(template, appDirectory);
    strcat(template, filename);
    {
        char * templateName = WWW_nameOfFile(template);
	templateAnchor = (HTParentAnchor*)HTAnchor_findAddress(templateName);
	free(templateName);
	HTRequest_clear(request);
	HTLoadAnchor((HTAnchor*)templateAnchor, request);
    }
	/* Cheat: Swap the loaded document to a different anchor */
    { 
        HyperText * doc = (HyperText*) HTAnchor_document(templateAnchor);
	HTAnchor_setDocument(myAnchor, (void*)doc);
	HTAnchor_setDocument(templateAnchor, NULL);
	[doc setNodeAnchor: myAnchor];
    }
    HTHistory_record((HTAnchor*)myAnchor);
    return myAnchor;
}




//	Go Home
//	-------
//
//	This accesses the default page of text for the user or, failing that,
//	for the system. 
//
static HTParentAnchor * home = NULL;
- goHome:sender
{
    if (!home) home = HTHomeAnchor();
    HTRequest_clear(request);
    if (! HTLoadAnchor((HTAnchor*)home, request)) return nil;
    /* HTHistory_record(myAnchor); */
    return self;
}


//	Load Help information
//	---------------------
//
//
- help:sender
{
    HTRequest_clear(request);
    if (! HTLoadAbsolute(HELP_ENTRY, request) ) return nil;
    HTHistory_record((HTAnchor*)HTMainAnchor);
    return self;
}


//	Go to the Blank Page
//	--------------------
//
//
- goToBlank:sender
{
    return [self openMy:"blank.html" diagnostic:0] ? self : nil;
}

//				Application Delegate Methods
//				============================


//	On Initialisation, Load Initial File
//	------------------------------------

-appDidInit:sender
{
    if (TRACE) printf("HyperManager: appDidInit\n");
    
//    StrAllocCopy(appDirectory, NXArgv[0]);
//    if (p = strrchr(appDirectory, '/')) p[1]=0;	/* Chop home slash */
//    if (TRACE) printf("WWW: Run from %s\n", appDirectory);
    
    return [self goHome:self];
}

//	Accept that we can open files from the workspace

- (BOOL)appAcceptsAnotherFile:sender
{
    return YES;
}

//	Open file from the Workspace
//
- (int)appOpenFile:(const char *)filename type:(const char *)aType
{
    char * name = WWW_nameOfFile(filename);
    HyperText * HT = [self accessName:name Diagnostic:0];
    free(name);
    return (HT!=0);
}

//	Open Temporary file
//
//	@@ Should unlink(2) the file when we have done with it!

- (int)appOpenTempFile:(const char *)filename type:(const char *)aType
{
    char * name = WWW_nameOfFile(filename);	/* No host */
    HyperText * HT = [self accessName:name Diagnostic:0];
    free(name);
    return (HT!=0);
}



//		Actions:
//		-------
- search:sender
{
    return [self searchDiagnostic:0];
}

- searchRTF:sender
{
    return [self searchDiagnostic:1];
}

- searchSGML:sender
{
    return [self searchDiagnostic:2];
}


//	Direct open buttons:

- open:sender
{
    return [self accessName:[openString stringValueAt:0] Diagnostic:0];
}

- linkToString:sender
{
    return (id) [THIS_TEXT linkSelTo:
        	HTAnchor_findAddress([openString stringValueAt:0])
	   ];
}

- insertImage:sender
{
    HTChildAnchor * source;
    source = HTAnchor_findChildAndLink(
	[THIS_TEXT nodeAnchor],				/* parent */
	0,						/* Id */
	[openString stringValueAt:0],			/* Addresss */
	0);
	    
    return (id) [THIS_TEXT replaceSelWithImage: source
		alternative: NULL aligned: NULL isMap: NO
	   ];
}


- openRTF:sender
{
 return [self accessName:[openString stringValueAt:0] Diagnostic:1];
}

- openSGML:sender
{
 return [self accessName:[openString stringValueAt:0] Diagnostic:2];
}


//	Save a Given document
//	---------------
- saveDocument:(HyperText *) HT
{
    if (YES) {
        HTParentAnchor * anchor = [HT nodeAnchor];
    	HTStream * s;
	HTStructured * ss;
	char * addr;

	request->anchor = anchor;
	
	s = HTSaveStream(request);
	if (!s) return nil;
	ss = HTMLGenerator(s);
	
	addr = HTAnchor_address((HTAnchor*)anchor); //	The WWW address
        if ([HT writeSGML:ss relativeTo:addr])
		[[HT window] setDocEdited:NO];
        free(addr);

    } else {
    	HTAlert("Can only save HTML format");
	return nil;
    }
#ifdef NOT_YET
    else if (format==WWW_RICHTEXT) [HT writeRichText:s];
    else if (format==WWW_PLAINTEXT || format==WWW_SOURCE) [HT writeText:s];
    else fprintf(stderr, "HT/File: Unknown format!\n");
#endif    

    return self;		/* OK */
}

//	Save a document
//	---------------
- save:sender
{
    return [self saveDocument: THIS_TEXT];
}    


//	Save all hypertexts back
//	-------------------------

- saveAll:sender
{
    List * windows = [NXApp windowList];
    id cv;
    int i;
    int n = [windows count];
    
    for(i=0; i<n ; i++){
	Window * w = [windows objectAt:i];
	if (cv=[w contentView])
	 if ([cv respondsTo:@selector(docView)])
	 if ([w isDocEdited]) {
		HyperText * HT = [[w contentView] docView];
		if ([self saveDocument:HT])
			[w setDocEdited: NO];
	}
    }

    return self;
}

//////////////////////////////////////////////////////////////
//
//	File oriented actions
//

//	Get Filename near a "hint" node
//	-------------------------------
//
// On entry,
//	hint		is a hypertext whose name is to be used as a hint for
//			the user when selecting the name.
// On exit,
//	returns		0 if no file selected
//			pointer to static string with filename if ok
//
const char * ask_name(HyperText * hint, HTFormat format)
{
    char * 		suggestion;
    char * 		slash;
    int			status;
    static SavePanel * 	save_panel;		/* Keep a Save panel alive  */
    
    if (!hint) return 0;
    if (!save_panel) {
        save_panel = [SavePanel new];	//	Keep between invocations
    }
    
    if (format==WWW_HTML) [save_panel setRequiredFileType: "html"];
    else if (format==WWW_RICHTEXT) [save_panel setRequiredFileType: "rtf"];
    else if (format==WWW_PLAINTEXT) [save_panel setRequiredFileType: "txt"];
    else if (format==WWW_POSTSCRIPT) [save_panel setRequiredFileType: "ps"];
    else [save_panel setRequiredFileType: ""];
    
    {
        char * addr = HTAnchor_address((HTAnchor*)[THIS_TEXT nodeAnchor]);
	suggestion = HTLocalName(addr);
	free(addr);
    }
    slash = strrchr(suggestion, '/');	//	Point to last slash
    if (slash) {
    	*slash++ = 0;			/* Separate directory and filename */
	status = [save_panel runModalForDirectory:suggestion file:slash];
    } else {
        if (TRACE) printf ("No slash in directory!!\n");
	status = [save_panel runModalForDirectory:"." file:suggestion];
    }
    free(suggestion);
    
    if (!status) {
    	if (TRACE) printf("Save cancelled.\n");
	return 0;
    }
    return [save_panel filename];
}

//	Save a particular format to a file -- no backup.

- save: (HyperText *)HT inFile:(const char *)filename format:(HTFormat)format
{
    if (format==WWW_HTML) {
	FILE * fp = fopen(filename, "w");
	HTStream * s = HTFWriter_new(fp, NO);
	HTStructured * st = HTMLGenerator(s);
    	char *saveName = WWW_nameOfFile(filename);	//	The WWW address
        [HT writeSGML:st relativeTo:saveName];
        free(saveName);
    }
    else{
	NXStream * s;				//	The file stream
	s = NXOpenMemory(NULL, 0, NX_WRITEONLY);

        if (format==WWW_RICHTEXT) [HT writeRichText:s];
        else if (format==WWW_PLAINTEXT || format==WWW_SOURCE)
	 	[HT writeText:s];
        else fprintf(stderr, "HT/File: Unknown format!\n"); 
	if (TRACE) printf("HT file: %li bytes in file `%s' format %s.\n",
	    NXTell(s), filename, HTAtom_name(format));
	    
	NXFlush(s);			/* Try to get over missing end */
	NXSaveToFile(s, filename);
	NXCloseMemory(s, NX_FREEBUFFER);
    }
    return self;
}



//	Save as a file	using save panel	(any format)
//	--------------------------------
// On entry,
//	format	Gives the file format to be used (see WWW.h)
//	hint	is a node to be used for a suggested name.
//
- saveIn:(HTFormat)format like:(HyperText *)hint
{  
    const char * filename;		//	The name of the file selected
    HyperText * HT;
    
    HT = THIS_TEXT;
    if (!HT) return HT;			//	No main window!
    
    filename = ask_name(HT, format);
    if (!filename) return nil;		//	Save clancelled.
    
    return [self save:HT inFile:filename format:format]; 
}

//	Save as an HTML file	using save panel
//	--------------------
//
- saveAs:sender
{
    return [self saveIn:WWW_HTML like:THIS_TEXT];
}

//	Save as an RTF file	using save panel
//	--------------------
//
- saveAsRichText:sender
{
    return [self saveIn:WWW_RICHTEXT like:THIS_TEXT];
}

//	Save as a plain text file	using save panel
//	-------------------------
//
- saveAsPlainText:sender
{
    return [self saveIn:WWW_PLAINTEXT like:THIS_TEXT];
}
///////////////// Opening files:

//	Open a file of a given name
//	---------------------------
//
//
// On exit,
//	Returns	If successful, the anchor of the node loaded or found.
//		If failed, nil.

- openFile:(const char *)filename diagnostic:(BOOL)diagnostic
{
    char * node_address = WWW_nameOfFile(filename);
    HTRequest_clear(request);
    HTLoadAbsolute(node_address, request);
    free(node_address);
    return self;
}


//	Ask the User for the name of a file
//	-----------------------------------
//

const char * existing_filename()
{
    HyperText * HT;			//	The new node
    char * suggestion = 0;
    char * slash;
    int	status;
    char * typelist = 0;		//	Any extension.
    static OpenPanel * openPanel=0;


//	Get The Filename from the User
    
    if (!openPanel) {
    	openPanel = [OpenPanel new];
    	[openPanel allowMultipleFiles:NO];
    }

    HT = THIS_TEXT;			// Hypertext in main window, if selected.
    if (HT) {
        char * addr = HTAnchor_address((HTAnchor*)[HT nodeAnchor]);
	suggestion = HTLocalName(addr);
	free(addr);
	slash = strrchr(suggestion, '/');	//	Point to last slash
	if (slash) {
	    *slash++ = 0;			/* Separate directory and filename */
	    status = [openPanel runModalForDirectory:suggestion
	    			file:""
				types:&typelist];
	    // (was: file:slash but this is silly as that is already open.)
	} else {
            if (TRACE) printf ("No slash in directory!!\n");
	    status = [openPanel runModalForDirectory:"."
	    	file:"" types:&typelist];
	    // (was: file:suggestion but this is silly as above)
	}
	free(suggestion);
    } else {
        status = [openPanel runModalForTypes:&typelist];
    }

    if (!status) {
    	if (TRACE) printf("Open cancelled\n");
	return NULL;
    }
    return [openPanel filename];
}


//	Open A File using the panel			openDiagnostic:
//	---------------------------
//
// On exit,
//	Returns	If successful, the anchor of the node loaded or found.
//		If failed, nil.

- openDiagnostic:(int)diagnostic
{
    
    const char * filename = existing_filename();

    if (!filename) {
    	if (TRACE) printf("Open cancelled\n");
	return nil;
    }

    return [self openFile:filename diagnostic:diagnostic];
}


//	Direct open buttons:

- openFile:sender
{
   return [self openDiagnostic:0];
}

- openFileRTF:sender
{
    return [self openDiagnostic:1];
}

- openFileSGML:sender
{
    return [self openDiagnostic:2];
}



//	Link to a given file
//	--------------------

- linkToFile:sender
{
    const char * filename = existing_filename();	// Ask for filename
    if (filename) {
    	char * node_address = WWW_nameOfFile(filename);
    	HTAnchor * a =  HTAnchor_findAddress(node_address);
	free(node_address);
	
	return [THIS_TEXT linkSelTo:a] ? self : nil;
    }
    return nil;
}


//	Make a new blank node named like the current node
//	-------------------------------------------------
//
// On exit,
//	Returns	If successful, the anchor of the node loaded or found.
//		If failed, nil.
- makeNew:sender
{
    id		status;
    HyperText * hint = THIS_TEXT;	//	The old hypertext
    char *	node_address;		//	of the new hypertext
    HTParentAnchor *	a;			//	The new anchor
    const char * filename;		//	The name of the file selected
    HyperText * HT;			//	The new hypertext
    
    if (!hint) return nil;		//	No main window!

    a = [self openMy:"blank.html" diagnostic:0];
    if (!a) return nil;			//	Couldn't find blank node
    
    HT = (HyperText*) HTAnchor_document(a);
    if (!HT) return HT;			//	No node?!
    
    filename = ask_name(hint, WWW_HTML);
    if (!filename) return nil;		//	Save cancelled.
    
    {			/* Switch doc from template anchor to real */
    	HTParentAnchor * newAnchor;
	node_address = WWW_nameOfFile(filename);
    	newAnchor = (HTParentAnchor*) HTAnchor_findAddress(node_address);
	HTAnchor_setDocument(newAnchor, (void*)HT);
	[HT setNodeAnchor:newAnchor];
	HTAnchor_setDocument(a, NULL);
    }

/*	Make a default title for the document from the file name:
**
** On entry,
**	node_address must be valid and have a colon in it
** On exit,
**	Title is set
**	This is destructive of the node_address string.
*/
    {
	char * pDir = 0;			/* Last Directory name */
        char * pTitle = strrchr(node_address, '/');	/* File name */

	if (pTitle) {
	    *pTitle++ = 0;			/* Chop in two */
	    pDir = strrchr(node_address, '/');
	    if (!pDir) pDir = strchr(node_address, ':');
	    pDir++;	/* After delimiter 921122 */
	} else {
	    pTitle  = strrchr(node_address, ':')+1;
	    pDir = "";
	}
	if (strrchr(pTitle, '.')) *strrchr(pTitle, '.') = 0;// Kill filetype */

	if (pDir) {
    	    char title[255];
	    sprintf(title, "%s -- %s", pTitle, pDir);
	    [HT setTitle:title];		/* Default title is filename */
	} else {
	    [HT setTitle:pTitle];    
	}
    }
    
    free(node_address);
    
    status = [self save:HT inFile:filename format:WWW_HTML]; 
    if (!status) return nil;		//	Save failed!
    
    [HT setEditable:YES];		// This is editable data.

    return self;
}


//	Make a new blank node named like the current node and link to it
//	----------------------------------------------------------------
//
- linkToNew:sender
{
    HyperText * HT = THIS_TEXT;
    HTParentAnchor * a;
    if (![HT isEditable]) return nil;	/* Won't be able to link to it */
    
    if (! [self makeNew:sender]) return nil;	/* Make the new node */
    a = [THIS_TEXT nodeAnchor];
    return [HT linkSelTo:(HTAnchor*)a] ? self : nil;
    	/* Link the selection to it */
}

//	Post to newsgroup etc
//
//	The newsgroup must be marked.
//
- postToMarked:sender
{
    HyperText * HT = THIS_TEXT;
    HTParentAnchor * source;
    HTAnchor * dest;

    if (![HT isEditable]) return nil;	/* Won't be able to link to it */
    
    dest = HTAnchor_findAddress([openString stringValueAt:0]);
    source = [THIS_TEXT nodeAnchor];
    HTAnchor_link((HTAnchor*)source, dest, HTAtom_for("posted-to"));
    return self;
}

/////////////////////////////////////////////////////////////////////

//	Close all unedited windows except this one
//	------------------------------------------
//

- closeOthers:sender
{
    Window * thisWindow = [NXApp mainWindow];
    List * windows = [[NXApp windowList] copy];

    {
        int i;
	id cv;					// Content view
	int n = [windows count];
        for(i=0; i<n; i++){
	    Window * w = [windows objectAt:i];
	    if (w != thisWindow)
	    if (cv=[w contentView])
	    if ([cv respondsTo:@selector(docView)]) {
	    	if (![w isDocEdited]) {
		    if (TRACE) printf(" Closing window %p\n", (void*)w);
		    [w performClose:self];
	        }
	    }
	}
	[windows free];				/* Free off copy of list */
	return self;
    }
}

//	Print Postscript code for the main window
//	-----------------------------------------

- print:sender
{
     return [THIS_TEXT printPSCode:sender];
}

//	Run the page layout panel
//
- runPagelayout:sender
{
    PageLayout * pl = [WWWPageLayout new];
    [pl runModal];
    return self;
}

//	Set the title of the main window
//	--------------------------------

- setDocTitle: sender
{
    Window * thisWindow = [NXApp mainWindow];
    [thisWindow setTitle:[titleString stringValueAt:0]];
    [thisWindow setDocEdited:YES];
    return self;
}

//	Inspect Link
//	------------

- inspectLink:sender
{
    HTChildAnchor * source = [THIS_TEXT selectedLink];
    HTAnchor * destination;
    if (!source){
    	[openString setStringValue: "(No anchor selected in main document.)"
                at:0];
         return nil;
    }
    {
    	char * source_address = HTAnchor_address((HTAnchor*)source);
    	[addressString setStringValue: source_address];
	free(source_address);
    }

    destination = HTAnchor_followMainLink((HTAnchor*)source);
    if (destination) {
    	char * destination_address = HTAnchor_address((HTAnchor*)destination);
    	[openString setStringValue: destination_address at:0];
	free(destination_address);
    } else {
	[openString setStringValue: "Anchor not linked."  at:0];
    }

    return self;
}

//	Inspect Image
//	------------

- inspectImage:sender
{
    HTChildAnchor * source = [THIS_TEXT selectedImageAnchor];
    HTAnchor * destination;
    if (!source){
    	[openString setStringValue: "(No image selected in main document.)"
                at:0];
         return nil;
    }
    {
    	char * source_address = HTAnchor_address((HTAnchor*)source);
    	[addressString setStringValue: source_address];
	free(source_address);
    }

    destination = HTAnchor_followMainLink((HTAnchor*)source);
    if (destination) {
    	char * destination_address = HTAnchor_address((HTAnchor*)destination);
    	[openString setStringValue: destination_address at:0];
	free(destination_address);
    } else {
	[openString setStringValue: "Image not linked."  at:0];
    }

    return self;
}

//	Mark address of document
//	------------------------
- copyAddress:sender
{
    char * addr = HTAnchor_address((HTAnchor*)[THIS_TEXT nodeAnchor]);
    [openString setStringValue:addr at:0];
    free(addr);
    return self;
}

//	Mark address of selection
//	-------------------------
- markSelected:sender
{
    char * addr = HTAnchor_address((HTAnchor*)[THIS_TEXT referenceSelected]);
    [openString setStringValue:addr at:0];
    free(addr);
    return self;
}

- unlink:sender
{
    return [THIS_TEXT unlinkSelection];
}


////////////////////////////////////////////////////////////////////
//	Old LINK menu commands from TextToy
//

#ifdef OLD_CODE

HTAnchor *	Mark;		/* A marked Anchor */

/*	Set up the start and end of a link
*/

- (HTChildAnchor*) linkToMark:sender		// Use linkToString
{
    /* return [THIS_TEXT linkSelTo:Mark]; */
}


- (HTParentAnchor*) markAll:sender		// USE copyAddress
{
    Mark = (HTAnchor*) [THIS_TEXT referenceAll];
    return (HTParentAnchor*) Mark;
}

#endif

/////////////////////////////////////////////////////////////////////////
//
//		HyperText delegate methods
//		==========================
//
//	This one has been passed from a window
//	to the hypertext which is its delegate,
//	to the access server module of that hypertext,
//	to this access manager.
//
// When a hypertext window becomes a key window, the search
// panel is turned on or off depending on whether a search can be done,
// and the default address in the "open using full reference" panel
// is set to the address of the current hypertext.
//
- hyperTextDidBecomeMain: sender
{

    HTParentAnchor * anchor = [(HText*)sender nodeAnchor];
    char * addr = HTAnchor_address((HTAnchor*)anchor);
    
    HTMainText = sender;		/* Record for HTAccess */
    HTMainAnchor = anchor;
    
    if (HTAnchor_isIndex([sender nodeAnchor])) {
        [[keywords window] makeKeyAndOrderFront:self];
    } else {
        [[keywords window] close];
//        [[keywords window] orderOut:self];	bug?
    }
    [titleString setStringValue: [[sender window] title] at:0];
    [addressString setStringValue: addr];
//  [openString setStringValue: addr at:0];
    free(addr);
    return self;
}

//	Panel delegate methods
//
//	The only windows to which this object is a delegate
//	are the open and search panels. When they become key,
//	we ensure that the text is selected.

- windowDidBecomeKey:sender
{
    if (sender == [openString window])
        [openString selectTextAt:0];	// Preselect the text
    else if (sender == [keywords window])
        [keywords selectTextAt:0];	// Preselect the text

    return self;
}
@end
