//  								ImageCell.m
//
//  		An IMG image within a hypertext object
//
//  (c) TBL CERN 1993
//
#import "ImageCell.h"

#define THEDIRECTIVE "WWWimg"

@implementation ImageCell


struct _HTStructured {		/* All we need to know */
	HTStructuredClass * isa;
	/* ... */
};

//sent when the class receives its first message
+ initialize
{
	[Text registerDirective:THEDIRECTIVE forClass:[ImageCell class]];
	return self;
}
- init
{
	self = [super init];

	image = nil;
	fileName = NULL;
	imageAnchor = NULL;
	linkAnchor = NULL;
	return self;
}


//	Apparatus to build the cell from the web
//	----------------------------------------

//	Don't use directly

- setFile:(const char *)theFile
{
#ifdef OLD_CODE
    image = nil;
    
    if(theFile != NULL){
	    fileName = NXCopyStringBuffer(theFile);
	    image = [[NXImage alloc] initFromFile:fileName];
    }
    if(image == nil) image = [NXImage findImageNamed:"NXdefaulticon"];
#else
    if(theFile != NULL) {
	image = [NXImage findImageNamed:theFile];
	if (image == nil) {
		fileName = NXCopyStringBuffer(theFile);
		image = [[NXImage alloc] initFromFile:fileName];
		[image setName:fileName];
	}
    }

    [self setIcon:theFile];
#endif
    return self;
}

static HTRequest * embedRequest = NULL;

PRIVATE BOOL embed_callback(HTRequest * request, void * filename)

{
    if (!request->context)	/* Only once */
	HTSACopy((char**)(&(request->context)), (char*)filename);
    return YES;
}

//	Assumes image info is set up

- initImage
{  
    HTAnchor * dest = HTAnchor_followMainLink((HTAnchor*) imageAnchor);
    char * filename;
    
    if (!embedRequest) {
        CONST char * CONST * image_suffixes = [NXImage imageFileTypes];
    	
	embedRequest = HTRequest_new();
	
	{
	    CONST char* CONST * p = image_suffixes;
	    HTAtom * octet_stream = HTAtom_for("application/octet-stream");
	    HTAtom * www_unknown = HTAtom_for("www/unknown");
	    HTAtom * www_present = HTAtom_for("www/present");
	    while (*p) {
	    	char dummy_filename[50];
	        HTFormat file_format;
		sprintf(dummy_filename, "fred.%s", *p);
		file_format = HTFileFormat(dummy_filename, NULL, NULL);
		if (TRACE) fprintf(stderr,
		"  Image suffixes include %s which means %s format\n",
			*p, file_format ? HTAtom_name(file_format)
					: "an unknown");
		if (file_format && file_format != www_unknown
			&& file_format != octet_stream) {
			
		    HTList * scan = embedRequest->conversions;
		    HTPresentation * p;
		    while (p = HTList_nextObject(scan)) {/* Check duplicate */
		        if ((p->rep == file_format)
			    && (p->rep_out == www_present)) break;
		    }
		    if (!p) HTSetConversion(embedRequest->conversions,
		    	HTAtom_name(file_format), "www/present",
			HTSaveAndCallBack,
			1.0, 0.0, 0.0);
		}
		p++;
	    }
	    
	    /* We can always take MIME */
	    HTSetConversion(embedRequest->conversions,
	     	"www/mime", "*", HTMIMEConvert,	1.0, 0.0, 0.0);
	}
    }
    
    HTRequest_clear(embedRequest); /* clear up any errors */
    embedRequest->callback = embed_callback;
    embedRequest->context = NULL;
    HTLoadAnchor(dest, embedRequest);
    filename = (char *)embedRequest->context;
    
    if (TRACE) fprintf(stderr, "ImageCell: File name is %s",
    	filename);
	
    [self setFile: filename ? filename : "docflash2.tiff"];
    
    if (filename) free(filename);
    
    return self;
}



//	Set up from SGML style instance variables
//	-----------------------------------------

- generateImageInfoFromAttributes
{
    char * href = NULL;
    if (present[HTML_IMG_SRC]  && value[HTML_IMG_SRC]) {
	StrAllocCopy(href, value[HTML_IMG_SRC]);
	HTSimplify(href);
    }
    imageAnchor = HTAnchor_findChildAndLink(
	[view nodeAnchor],				/* parent */
	0,						/* Id */
	present[HTML_IMG_SRC] ? href : 0,		/* Addresss */
	0);

    return self;
}


//	Image-specific init method
//
//	We generate a relative link from the document.

-initAnchor: (HTChildAnchor *) anc
	alternative: (CONST char *) thisAlt
	aligned: (CONST char *) thisAlignment
	isMap:   (BOOL)		thisIsMap
	inView:	(HyperText*)	v
{
    int attributes = HTML_IMG_ATTRIBUTES;
    HTAnchor * destination;

    [self init];
    
    view = v;
    imageAnchor = anc;

    
    dtd = &HTMLP_dtd;
    element_number = HTML_IMG;
   
    present = (BOOL*)calloc(attributes, sizeof(BOOL));
    value = (char**) calloc(attributes, sizeof(char*));
    
    if (thisAlt) {
        StrAllocCopy(value[HTML_IMG_ALT], thisAlt);
	present[HTML_IMG_ALT] = YES;
    }
    
    if (thisAlignment) {
        StrAllocCopy(value[HTML_IMG_ALIGN], thisAlignment);
	present[HTML_IMG_ALIGN] = YES;
    }
    
    present[HTML_IMG_ISMAP] = thisIsMap;
    present[HTML_IMG_SRC] = YES;
    destination = HTAnchor_followMainLink((HTAnchor*)imageAnchor);
    value [HTML_IMG_SRC] = HTAnchor_address(destination);

    [self initImage];
    return self;
}

//	Normal method -- given SGML attributes -- called from super init::::

- setup
{
     [self generateImageInfoFromAttributes];
     return [self initImage];
}


//	Writing SGML back involves relative URLs.
//	-----------------------------------------
//
//  We have to temporarily switch the URI for a relative one
//  for writing out the SGML.  We don't keep it in relative form
//  or we wouldn't be able to cut and paste them.
//
- writeSGML:(HTStructured *) thisTarget relativeTo:(const char *)aName

{
    char * absolute = value[HTML_IMG_SRC];
    char * document_uri = HTAnchor_address((HTAnchor*) [view nodeAnchor]);
    value [HTML_IMG_SRC] = HTRelative(absolute, document_uri);
    
    [super writeSGML:thisTarget relativeTo:aName];
    
    free(value[HTML_IMG_SRC]);
    value[HTML_IMG_SRC] = absolute;
    free(document_uri);
    return self;
}


//	Free
//	----

- free
{
#ifdef OLD_CODE
	id tmpImage= [NXImage findImageNamed:"NXdefaulticon"];
	
	if((image != nil) && (image != tmpImage)) [image free];
	if(fileName != NULL) NX_FREE(fileName);
#else
	if (fileName) {
	    NX_FREE (fileName);
//	    [image free];	// @@@ Need an image reference count !
	}
#endif
	return [super free];
}


- (HTChildAnchor*) imageAnchor  	{ return imageAnchor; }



#ifdef OLD_CODE
- calcCellSize:(NXSize *)theSize
{
	if(image != nil) [image getSize:theSize];
	return self;
}

- drawSelf:(const NXRect *)rect inView:view
{
	NXPoint tmpPoint = {0.0,0.0};
	
	[view getLocation:&tmpPoint ofCell:self];//view is the Text object
	
	tmpPoint.y += (rect->size).height;//the coordinates are flipped
	
	[image composite:NX_SOVER toPoint:&tmpPoint];
	return self;
}

- highlight:(const NXRect *)rect inView:view lit:(BOOL)flag
{
	return self;
}
#endif



@end
