/*				FORMAT CONVERSION FROM SGML
**				===========================
**
**
**	22 Nov 92	Fixed quoting of hrefs.
**			CERN_WEIRDO ifdefed out -- proper SGML expected
**			REMOVE_SCRIPT ifdefed out -- did ignore lines starting with "."
*/

#import  "HTStyle.h"
#include "HTML.h"		/* For directory object building */
#include "SGML.h"		/* Structured stream def */

struct _HTStructured {
	HTStructuredClass * isa;
	/* ... */
};

#define PUTC(c) (*targetClass.put_character)(target, c)
#define PUTS(s) (*targetClass.put_string)(target, s)
#define START(e) (*targetClass.start_element)(target, e, 0, 0)
#define END(e) (*targetClass.end_element)(target, e)


struct _HTStream {
	CONST HTStreamClass *	isa;
	/* ... */
};

/*	HTML2
*/
extern void HTRegenInit NOPARAMS;

extern void HTRegenCharacter PARAMS((
	char			c,
	HTStyle *		style,
	HTStructured *		target));

extern	void HTNestingChange PARAMS(
	(HTStructured* 	s, 
	HTStyle*		old, 
	HTStyle *		new));


/*		Write SGML File back OUT
**		------------------------
**
**	This is currently quite NeXT-specific.
**
**	We run through te runs. When a characteristic of a run changes, we
**	output the approporiate SGML code. When several characteristics change at
**	the same place, we output the code in an order such that the resulting
**	structures wil be nested. This means first unwrapping the old ones, and
**	then entering the new ones. For example, it is better to produce
**
**		<h2><a>...</a></h2><a>...</a>
**	than
**
**		<h2><a>...</h2></a><a>...</a>
**
**	The special treatment of newlines is because we want to strip extra newlines
**	out. We ignore newlines at the beginning and end of the para style,
**	and we treat multiple newlines as a single paragraph mark.
**
**	Bugs:	@@@ Highlighting is ignored.
**		@@@ end text is ignored.
*/

static SGML_gen_errors;		/* Number of unrcognizable runs */
static const char * saveName;	/* pointer to name node is being saved under */
static int	lineLength;	/* Number of characters on a line so far */


/*	Change Run
**	==========
*/
/*	This function generates the code for one run, given the previous run.
**
*/
- (void) changeRunFrom: (NXRun *) last to: (NXRun *) r
{

   int chars_left = r->chars;
    
    HTStyle* old_style = (HTStyle*) last->paraStyle;
    HTStyle* new_style = (HTStyle*) r->paraStyle;
    
    HTNestingChange(target, old_style, new_style);

#ifdef JUNKELWUB

    if (r->info != last->info) {			/* End anchor */
	if (last->info) END(HTML_A);
    }
    
    if (r->paraStyle != last->paraStyle)
     if (last->paraStyle) {				/* End paragraph */
	if (currentSGML) PUTS(currentSGML->end_tag);
	else START(HTML_P);
	lineLength = 0;	 /* At column 1 */
    }

    
    if (r->paraStyle != last->paraStyle) {		/* Start paragraph */
	currentSGML = findSGML(r->paraStyle);
	if (currentSGML) {

	    if (currentSGML->free_format)
	     while(chars_left && WHITE(*read_pointer)) {/* Strip leading */
		(chars_left)--;				/*   white space */
		(void) NEXT_TEXT_CHAR;
	    }
	    PUTS(currentSGML->start_tag);
	    prefix = currentSGML->start_text;
	} 
	SGML_gen_newlines=0;				/* Cancel  */
    }
    
    if (r->info != last->info) {			/* Start anchor */

	if (SGML_gen_newlines) {	/* Got anchor, need paragraph separator */
	    PUTS(currentSGML->paragraph_tag);
	    SGML_gen_newlines=0;	/* paragraph flushed. */
	}
	if (r->info) {
	    HTChildAnchor * a = (HTChildAnchor *) r->info;
	    HTAnchor * d = HTAnchor_followMainLink((HTAnchor*)a);
	    char * this = HTAnchor_address((HTAnchor*)a);
	    char * that = NULL;

	    HTStartAnchor(target,
	    	strrchr(this, '#')+1,
	    	d ? that=HTAnchor_address((HTAnchor*)d) : NULL);
	    free(this);
	    if (that) free(that);
	}
    }
#endif

/*	Now output the textual part of the run
*/

    while (chars_left) {	
	char c = NEXT_TEXT_CHAR;
	chars_left--;
	HTRegenCharacter(c, new_style, target);
    }
    
} /* changeRunFrom:to: */



/*	This is the body of the SGML output method.
*/
- writeSGML:(HTStructured *) thisTarget relativeTo:(const char *)aName
{
    NXRun * r = theRuns->runs;
    int sor;				/* Character position of start of run */
    NXRun dummy;    

    dummy.paraStyle = 0;
    dummy.info = 0;
    dummy.chars = 0;
    
    target = thisTarget;	/* save in instance variable */
    targetClass = *target->isa; /* copy access routines */

    HTRegenInit();		/* Initialize regeneration of HTML */    
    HT = self;
    saveName = aName;

    SGML_gen_errors = 0;
    
    START_INPUT;
    lineLength = 0;			/* Starting in column 1 */
    		
    START(HTML_HTML);
    START(HTML_HEAD);
    START(HTML_TITLE);
    PUTS([window title]);
    END(HTML_TITLE);
    
    if (nextAnchorNumber) HTNextID(target, nextAnchorNumber);
    
    END(HTML_HEAD);
    START(HTML_BODY);

/*	Change style tags etc
*/
    [self changeRunFrom:&dummy to:r];			/* Start first run */
    
    for (sor=r++->chars; sor<textLength; sor=sor+(r++)->chars)  {
        if (TRACE) printf("%4i:  %i chars in run %3li.\n",
			sor, r->chars, r-theRuns->runs);
	[self changeRunFrom:r-1 to: r];	/* Runs 2 to N */
    }
    [self changeRunFrom:r to:&dummy];			/* Close last run */

    tFlags.changeState = 0; 		/* Please notify delegate if changed */
    END(HTML_BODY);
    END(HTML_HTML);
    target->isa->free(target);
    
    return (SGML_gen_errors) ? nil : self;    
}
